/*
 * libcache
 *
 * Author: Matt Miller
 *		(mmiller@hick.org)
 *
 * String parsing library 
 *
 */

#ifndef _LCACHE_H
#define _LCACHE_H

#define CACHE_SIZE 	64
#define BUF_SIZE	512	

#define A_READLNP	0x01
#define A_READLN	0x02
#define A_READP		0x03
#define A_READ		0x04

#define E_ERROR			 0
#define E_FD_CLOSED		-1
#define E_CACHE_BUFFER_FULL	-2
#define E_CACHE_FULL		-3
#define E_BAD_CACHE		-4
#define E_BAD_ACTION		-5
#define E_BAD_WORD		-6

#define S_OK		0x01
#define S_GOTEOLN	0x02

typedef struct cache_template {
        char *str[CACHE_SIZE]; 

        int words;
        char buf[BUF_SIZE];
        int bufpos;        
} CACHE;                 

#ifdef __cplusplus		/* Added for C++ compatability, thanks Mahatios ap Ganadhor */
extern "C" {
#endif

int cacheParseWords(CACHE *cache);
int cacheFree(CACHE *cache);
int cacheInitialize(CACHE *cache);
int cacheReadFileFDLine(FILE *fd, CACHE *cache);
int cacheReadSockFD(int fd, CACHE *cache, int action);
int cacheAddCharacter(CACHE *cache, char character);
int _cacheReadSock(int fd, CACHE *cache, int action);
void cacheCompileWords(CACHE *cache, int start, int finish, char *delim,char *retbuf,unsigned int retbufsize);

#ifdef __cplusplus
}
#endif

#endif
