/*
 *              http://www.afro-productions.com
 * sc   12/2001         mmiller@hick.org        Matt Miller
 *
 *                          SLAPChat
 *            Secure Layer (Afro Production) Chat
 *                       (c) 2001-2002
 *			All rights reserved.
 *
 * All code below and above is released freely to the public with only two
 * limitations:
 *      a) Any desired use of this code for a for-profit nature requires express
 *         permission from the author.
 *      b) This code can be used to serve whatever purpose is desired.  With
 *         that in mind, the original author cannot be held liable for what this
 *         code is used for and therefore will NOT be held liable for the actions
 *         of others.  As always, give credit where credit is due.
 *
 * Further information about licensing and liability can be found in the file
 * 'COVER' that should have been included with this file.  If it was not,
 * please contact afro productions and request a copy.
 */

#ifndef _SC_COMMON_HASH_H
#define _SC_COMMON_HASH_H

#ifdef __cplusplus
extern "C" {
#endif

/*
 * hash related defines
 */

#define DEFAULT_HASH_BUCKETS		8192
#define DEFAULT_HASH_BUCKETS_MASK	(DEFAULT_HASH_BUCKETS-1)

/*
 * hash related structs 
 *      these should only be used with string or int keys.
 */

typedef struct hash_entry {
	unsigned int id;
	void *ptr;

	struct hash_entry *next;
} HASH_ENTRY;

typedef struct hash_table {
	HASH_ENTRY *hashEntry;
} HASH_TABLE;

/*
 * all hashes, whether they're defined as static or dynamic, are actually
 * dynamic.
 */

typedef struct hash_dynamic {
	unsigned long flags;
	unsigned int hashBuckets;

	HASH_TABLE *hashTable;
} DYNAMIC_HASH;

#define HASH_FLAG_ISDYNAMIC	(1 << 1)
#define HASH_FLAG_ISSTATIC	(1 << 2)
#define HASH_FLAG_FREEVALUE	(1 << 3)

#define HASH_GROUP		DYNAMIC_HASH	/* so we can operate on both dynamic and static hashes. */
#define STATIC_HASH		DYNAMIC_HASH	/* for readability.					*/

void hashInitialize(HASH_GROUP *hg, unsigned long flags, unsigned int hashBuckets);
void hashDeinitialize(HASH_GROUP *hg);

int hashStringKeyAdd(HASH_GROUP *hg, const char *key, void *value);	/* add a value to a given hash table 		*/
int hashStringKeyDelete(HASH_GROUP *hg, const char *key);		/* remove a value from a given hash table 	*/

void *hashStringKeyGetValue(HASH_GROUP *hg, const char *key);

unsigned int _hashStringKeyGetPos(HASH_GROUP *hg, const char *key, unsigned int *id, int dmax);
HASH_ENTRY *_hashStringKeyGetEntry(HASH_GROUP *hg, const char *key);

int hashIntegerKeyAdd(HASH_GROUP *hg, int key, void *value);
int hashIntegerKeyDelete(HASH_GROUP *hg, int key);

void *hashIntegerKeyGetValue(HASH_GROUP *hg, int key);

unsigned int _hashIntegerKeyGetPos(HASH_GROUP *hg, int key, unsigned int *id, int dmax);
HASH_ENTRY *_hashIntegerKeyGetEntry(HASH_GROUP *hg, int key);

#ifdef __cplusplus
}
#endif

#endif
