/*
 *              http://www.afro-productions.com
 * sc   12/2001         mmiller@hick.org        Matt Miller
 *
 *                          SLAPChat
 *            Secure Layer (Afro Production) Chat
 *                       (c) 2001-2002
 *			All rights reserved.
 *
 * All code below and above is released freely to the public with only two
 * limitations:
 *      a) Any desired use of this code for a for-profit nature requires express
 *         permission from the author.
 *      b) This code can be used to serve whatever purpose is desired.  With
 *         that in mind, the original author cannot be held liable for what this
 *         code is used for and therefore will NOT be held liable for the actions
 *         of others.  As always, give credit where credit is due.
 *
 * Further information about licensing and liability can be found in the file
 * 'COVER' that should have been included with this file.  If it was not,
 * please contact afro productions and request a copy.
 */

#ifndef _SC_COMMON_LIST_H
#define _SC_COMMON_LIST_H

#ifdef __cplusplus
extern "C" {
#endif

typedef struct linked_list_item {
	union {
		long longId;
		const char *stringId;
	} id;
	void *value;

	struct linked_list_item *prev;	
	struct linked_list_item *next;
} LINKED_LIST_ITEM;

typedef struct linked_list {
	unsigned int items;
	unsigned int flags;

	LINKED_LIST_ITEM *list;
} LINKED_LIST;

#define LIST_FLAG_FREEVALUE	(1 << 1)
#define LIST_FLAG_THREADSAFE	(1 << 2)
#define LIST_FLAG_NOTGENOBJ	(1 << 3)
#define LIST_FLAG_APPENDEND	(1 << 4)

/*
 * id independent functions
 */

int listInitialize(LINKED_LIST *list,unsigned int flags);
int listDeinitialize(LINKED_LIST *list);

void listFlush(LINKED_LIST *list);
inline LINKED_LIST_ITEM *listAddItem(LINKED_LIST *list, void *value);
inline int listDeleteItem(LINKED_LIST *list, LINKED_LIST_ITEM *item);

/*
 * string keyed list functions
 */

int listAddStringId(LINKED_LIST *list, const char *key, void *value);
int listAddUniqueStringId(LINKED_LIST *list, const char *key, void *value);

int listDeleteStringId(LINKED_LIST *list, const char *key);

LINKED_LIST_ITEM *listGetItemStringId(LINKED_LIST *list, const char *key);
void *listGetValueStringId(LINKED_LIST *list, const char *key);

short listHasStringId(LINKED_LIST *list, const char *key);

/*
 * long keyed list functions
 */

int listAddLongId(LINKED_LIST *list, long key, void *value);
int listAddUniqueLongId(LINKED_LIST *list, long key, void *value);

int listDeleteLongId(LINKED_LIST *list, long key);

LINKED_LIST_ITEM *listGetItemLongId(LINKED_LIST *list, long key);
void *listGetValueLongId(LINKED_LIST *list, long key);

short listHasLongId(LINKED_LIST *list, long key);

/*
 * string and integer keyed list functions
 */

int listAddStringAndLongId(LINKED_LIST *list, const char *stringId, long longId, void *value);
int listAddUniqueStringAndLongId(LINKED_LIST *list, const char *stringId, long longId, void *value);

int listDeleteStringAndLongId(LINKED_LIST *list, const char *stringId, long longId);

LINKED_LIST_ITEM *listGetItemStringAndLongId(LINKED_LIST *list, const char *stringId, long longId);
void *listGetValueStringAndLongId(LINKED_LIST *list, const char *stringId, long longId);

/*
 * list iterator func --> these are not threadsafe, make sure you lock and
 * unlock accordingly.
 */

LINKED_LIST_ITEM *listGetFirstItem(LINKED_LIST *list);
LINKED_LIST_ITEM *listGetPreviousItem(LINKED_LIST_ITEM *item);
LINKED_LIST_ITEM *listGetNextItem(LINKED_LIST_ITEM *item);

/*
 * this is the proper function to get an items value
 */

void *listGetItemValue(LINKED_LIST_ITEM *item);

#ifdef __cplusplus
}
#endif

#endif
