#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <time.h>

#include "won.h"

typedef struct _log_module_extension {
	char 		*logDirectory;

	LINKED_LIST	logables;
} LOG_MODULE_EXTENSION;

uint32 configLogDirectory(WON_CONFIG_PACKET *packet);

uint32 ircPrivmsg(WON_IRC_PACKET *wip);
uint32 ircJoin(WON_IRC_PACKET *wip);
uint32 ircPart(WON_IRC_PACKET *wip);
uint32 ircMode(WON_IRC_PACKET *wip);
uint32 ircKick(WON_IRC_PACKET *wip);
uint32 ircTopic(WON_IRC_PACKET *wip);

uint32 lowlevelSend(WON_MODULE_CTX *modctx, WON_PACKET *packet);

FILE *logOpenIfNot(LOG_MODULE_EXTENSION *extension,const char *ref);

uint32 wonLoad(WON_MODULE_CTX *modctx)
{
	LOG_MODULE_EXTENSION *extension = wonMemoryAllocate(sizeof(LOG_MODULE_EXTENSION),1);

	wonIrcRegisterLayeredHandler(modctx->wonctx, modctx, WON_LAYER_LAST, "PRIVMSG", 1, ircPrivmsg);
	wonIrcRegisterLayeredHandler(modctx->wonctx, modctx, WON_LAYER_LAST, "JOIN", 1, ircJoin);
	wonIrcRegisterLayeredHandler(modctx->wonctx, modctx, WON_LAYER_LAST, "PART", 1, ircPart);
	wonIrcRegisterLayeredHandler(modctx->wonctx, modctx, WON_LAYER_LAST, "KICK", 1, ircKick);
	wonIrcRegisterLayeredHandler(modctx->wonctx, modctx, WON_LAYER_LAST, "MODE", 1, ircMode);
	wonIrcRegisterLayeredHandler(modctx->wonctx, modctx, WON_LAYER_LAST, "TOPIC", 1, ircTopic);

	wonConfigRegisterLayeredHandler(modctx->wonctx, modctx, WON_LAYER_LAST, "log_directory", configLogDirectory);

	wonLowlevelSendRegisterLayeredHandler(modctx->wonctx, modctx, WON_LAYER_FIRST, lowlevelSend);

	listInitialize(&extension->logables,0);

	modctx->moduleExtension = extension;

	return WON_STATUS_SUCCESS;
}

uint32 wonIoctl(WON_MODULE_CTX *modctx,uint32 ioctl, WON_IC *wic)
{
	LOG_MODULE_EXTENSION *extension = (LOG_MODULE_EXTENSION *)modctx->moduleExtension;

	switch (ioctl)
	{
		case WON_IOCTL_LOAD_COMPLETE:
			if (!extension->logDirectory)
			{
				wonLog(modctx->wonctx,modctx,"no logging directory specified, unloading log.so");
				wonModuleDeferredUnload(modctx->wonctx,modctx->moduleName);
			}
		case WON_IOCTL_CONFIG_SAVE:
			if (wic->inLen == 4 && extension->logDirectory)
			{
				FILE *fd = (FILE *)wic->inBuffer;

				fprintf(fd,"# log.so configuration\n");
				fprintf(fd,"log_directory %s\n",extension->logDirectory);
			}
			break;
	}
	
	return WON_STATUS_SUCCESS;
}

uint32 wonUnload(WON_MODULE_CTX *modctx)
{
	LOG_MODULE_EXTENSION *extension = (LOG_MODULE_EXTENSION *)modctx->moduleExtension;
	LINKED_LIST_ITEM *item;

	wonIrcDeregisterLayeredHandler(modctx->wonctx, modctx, "PRIVMSG", 1, ircPrivmsg);
	wonIrcDeregisterLayeredHandler(modctx->wonctx, modctx, "JOIN", 1, ircJoin);
	wonIrcDeregisterLayeredHandler(modctx->wonctx, modctx, "PART", 1, ircPart);
	wonIrcDeregisterLayeredHandler(modctx->wonctx, modctx, "KICK", 1, ircKick);
	wonIrcDeregisterLayeredHandler(modctx->wonctx, modctx, "MODE", 1, ircMode);
	wonIrcDeregisterLayeredHandler(modctx->wonctx, modctx, "TOPIC", 1, ircTopic);
	
	wonLowlevelSendDeregisterLayeredHandler(modctx->wonctx, lowlevelSend);
	
	wonConfigDeregisterLayeredHandler(modctx->wonctx, modctx, "log_directory", configLogDirectory);

	for (item = listGetFirstItem(&extension->logables);
		item && item->value;
		item = listGetNextItem(item))
	{
		wonMemoryFree((char *)item->id.stringId);
		fclose((FILE *)item->value);
	}

	listDeinitialize(&extension->logables);

	return WON_STATUS_SUCCESS;
}

uint32 configLogDirectory(WON_CONFIG_PACKET *wcp)
{
	LOG_MODULE_EXTENSION *extension = (LOG_MODULE_EXTENSION *)wcp->modctx->moduleExtension;

	if (wcp->buffer->words == 1)
	{
		fprintf(stdout,"log: log_directory missing arguments.\n");
		return WON_STATUS_FAILURE;
	}

	wonMemoryFree(extension->logDirectory);

	extension->logDirectory = strdup(wcp->buffer->str[1]);
	
	return WON_STATUS_SUCCESS;
}

uint32 ircPrivmsg(WON_IRC_PACKET *wip)
{
	LOG_MODULE_EXTENSION *extension = (LOG_MODULE_EXTENSION *)wip->modctx->moduleExtension;
	char reason[1024];
	FILE *fd;

	if (wip->buffer->words < 4)
		return WON_STATUS_SUCCESS;

	fd = logOpenIfNot(extension,wip->buffer->str[2]);

	if (!fd)
		return WON_STATUS_SUCCESS;

	if (wip->buffer->str[3][1] != 1)
	{
		reason[1023] = 0;
		cacheCompileWords(wip->buffer,3,wip->buffer->words," ",reason,1023);
		 
		fprintf(fd,"%lu <%s> %s\n",(uint32)time(NULL),wip->fromNick,reason+1); 
		fflush(fd);
	}
	else if (!strcmp(wip->buffer->str[3]+2,"ACTION"))
	{
		reason[1023] = 0;
		cacheCompileWords(wip->buffer,4,wip->buffer->words," ",reason,1023);
		reason[strlen(reason)-1] = 0;
		 
		fprintf(fd,"%lu * %s %s\n",(uint32)time(NULL),wip->fromNick,reason); 
		fflush(fd);
	}

	return WON_STATUS_SUCCESS;
}

uint32 ircJoin(WON_IRC_PACKET *wip)
{
	LOG_MODULE_EXTENSION *extension = (LOG_MODULE_EXTENSION *)wip->modctx->moduleExtension;
	FILE *fd;

	if (wip->buffer->words < 3)
		return WON_STATUS_SUCCESS;

	fd = logOpenIfNot(extension,wip->buffer->str[2]+1);

	if (!fd)
		return WON_STATUS_SUCCESS;

	fprintf(fd,"%lu *** Join %s (%s)\n",(uint32)time(NULL),wip->fromNick,wip->fromMask);
	fflush(fd);

	return WON_STATUS_SUCCESS;
}

uint32 ircPart(WON_IRC_PACKET *wip)
{
	LOG_MODULE_EXTENSION *extension = (LOG_MODULE_EXTENSION *)wip->modctx->moduleExtension;
	FILE *fd;
	
	if (wip->buffer->words < 3)
		return WON_STATUS_SUCCESS;

	fd = logOpenIfNot(extension,wip->buffer->str[2]);

	if (!fd)
		return WON_STATUS_SUCCESS;

	fprintf(fd,"%lu *** Part %s (%s)\n",(uint32)time(NULL),wip->fromNick,wip->fromMask);
	fflush(fd);

	return WON_STATUS_SUCCESS;
}

uint32 ircMode(WON_IRC_PACKET *wip)
{
	LOG_MODULE_EXTENSION *extension = (LOG_MODULE_EXTENSION *)wip->modctx->moduleExtension;
	FILE *fd;
	char buf[1024];

	if (wip->buffer->words < 4)
		return WON_STATUS_SUCCESS;

	fd = logOpenIfNot(extension,wip->buffer->str[2]);

	if (!fd)
		return WON_STATUS_SUCCESS;

	buf[1023] = 0;
	cacheCompileWords(wip->buffer,3,wip->buffer->words," ",buf,1023);

	fprintf(fd,"%lu *** %s sets mode: %s\n",(uint32)time(NULL),wip->fromNick,buf);
	fflush(fd);

	return WON_STATUS_SUCCESS;
}

uint32 ircKick(WON_IRC_PACKET *wip)
{
	LOG_MODULE_EXTENSION *extension = (LOG_MODULE_EXTENSION *)wip->modctx->moduleExtension;
	FILE *fd;
	char buf[1024];

	if (wip->buffer->words < 5)
		return WON_STATUS_FAILURE;

	fd = logOpenIfNot(extension,wip->buffer->str[2]);

	if (!fd)
		return WON_STATUS_SUCCESS;

	buf[1023] = 0;
	cacheCompileWords(wip->buffer,4,wip->buffer->words," ",buf,1023);

	fprintf(fd,"%lu *** %s kicked %s (%s)\n",(uint32)time(NULL),wip->fromNick,wip->buffer->str[3],buf+1);
	fflush(fd);

	return WON_STATUS_SUCCESS;
}

uint32 ircTopic(WON_IRC_PACKET *wip)
{
	LOG_MODULE_EXTENSION *extension = (LOG_MODULE_EXTENSION *)wip->modctx->moduleExtension;
	FILE *fd;
	char buf[1024];

	if (wip->buffer->words < 4)
		return WON_STATUS_FAILURE;

	fd = logOpenIfNot(extension,wip->buffer->str[2]);

	if (!fd)
		return WON_STATUS_SUCCESS;

	buf[1023] = 0;
	cacheCompileWords(wip->buffer,3,wip->buffer->words," ",buf,1023);

	fprintf(fd,"%lu *** %s changes the topic to '%s' \n",(uint32)time(NULL),wip->fromNick,buf+1);
	fflush(fd);

	return WON_STATUS_SUCCESS;
}

FILE *logOpenIfNot(LOG_MODULE_EXTENSION *extension,const char *ref)
{
	FILE *fd = listGetValueStringId(&extension->logables,ref);
	char fileName[1024];
	char *newref;

	if (fd)
		return fd;

	fileName[1023] = 0;
	snprintf(fileName,1023,"%s/%s.log",extension->logDirectory,ref);

	fd = fopen(fileName,"a");

	if (!fd)
		return NULL;

	newref = strdup(ref);

	listAddStringId(&extension->logables,newref,fd);

	return fd;
}

uint32 lowlevelSend(WON_MODULE_CTX *modctx,WON_PACKET *packet)
{
	LOG_MODULE_EXTENSION *extension = (LOG_MODULE_EXTENSION *)modctx->moduleExtension;

	if (packet->type == WON_CONNECTION_DCC)
		return WON_STATUS_SUCCESS;

	if (packet->pkt.irc->buffer->words < 3)
		return WON_STATUS_SUCCESS;

	if (!strcmp(packet->pkt.irc->buffer->str[0],"PRIVMSG"))
	{
		FILE *fd = logOpenIfNot(extension,packet->pkt.irc->buffer->str[1]);
		char buf[1024];

		if (!fd)
			return WON_STATUS_SUCCESS;

		buf[1023] = 0;
		cacheCompileWords(packet->pkt.irc->buffer,2,packet->pkt.irc->buffer->words," ",buf,1023);

		fprintf(fd,"%lu <%s> %s\n",(uint32)time(NULL),"me",buf+1);
		fflush(fd);
	}

	return WON_STATUS_SUCCESS;
}
