#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>

#include "won.h"

typedef struct _quote_mod_extension {
	char		*quoteFile;

	LINKED_LIST	quotes;
} QUOTE_MOD_EXTENSION;

uint32 configQuoteFile(WON_CONFIG_PACKET *wcp);

uint32 ircPrivmsg(WON_IRC_PACKET *wip);

void cmdQuote(WON_IRC_PACKET *wip);
void cmdHelp(WON_IRC_PACKET *wip);
void cmdGet(WON_IRC_PACKET *wip);
void cmdAfroHug(WON_IRC_PACKET *wip);
void cmdAddQuote(WON_IRC_PACKET *wip);
void cmdSearch(WON_IRC_PACKET *wip);

void printQuote(QUOTE_MOD_EXTENSION *extension, WON_CTX *ctx,WON_IRC_CONNECTION *conn,const char *target,unsigned long quoteNumber);

void wonLoad(WON_MODULE_CTX *modctx)
{
	QUOTE_MOD_EXTENSION *extension;

	wonConfigRegisterLayeredHandler(modctx->wonctx,modctx, WON_LAYER_LAST, "quote_file",configQuoteFile);	

	wonIrcRegisterLayeredHandler(modctx->wonctx, modctx, WON_LAYER_LAST, "PRIVMSG", 1, ircPrivmsg);

	modctx->moduleExtension = wonMemoryAllocate(sizeof(QUOTE_MOD_EXTENSION),1);
	extension = (QUOTE_MOD_EXTENSION *)modctx->moduleExtension;

	listInitialize(&extension->quotes,LIST_FLAG_FREEVALUE|LIST_FLAG_APPENDEND);
}

void wonIoctl(WON_MODULE_CTX *modctx, uint32 ioctl, WON_IC *wic)
{
	switch (ioctl)
	{
		case WON_IOCTL_CONFIG_SAVE:
			if (wic->inLen == 4)
			{
				QUOTE_MOD_EXTENSION *extension = (QUOTE_MOD_EXTENSION *)modctx->moduleExtension;
	
				FILE *fd = (FILE *)wic->inBuffer;
				fprintf(fd,"# quote.so configuration\n");
	
				if (extension && extension->quoteFile)
					fprintf(fd,"quote_file %s\n",extension->quoteFile);
			}
			break;
	}
}

void wonUnload(WON_MODULE_CTX *modctx)
{
	QUOTE_MOD_EXTENSION *extension = (QUOTE_MOD_EXTENSION *)modctx->moduleExtension;

	listFlush(&extension->quotes);

	wonMemoryFree(extension->quoteFile);
	wonMemoryFree(extension);

	wonConfigDeregisterLayeredHandler(modctx->wonctx,modctx, "quote_file",configQuoteFile);	

	wonIrcDeregisterLayeredHandler(modctx->wonctx, modctx, "PRIVMSG", 1, ircPrivmsg);
}

uint32 configQuoteFile(WON_CONFIG_PACKET *wcp)
{
	QUOTE_MOD_EXTENSION *extension = (QUOTE_MOD_EXTENSION *)wcp->modctx->moduleExtension;
	char buf[1024];
	FILE *fd;

	if (wcp->buffer->words < 2)
	{
		fprintf(stdout,"quote: quote_file is missing arguments\n");
		return WON_STATUS_FAILURE;
	}

	if (!(fd = fopen(wcp->buffer->str[1],"r")))
	{
		fprintf(stdout,"quote: the quote_file specified was not valid.\n");
		return WON_STATUS_FAILURE;
	}

	buf[1023] = 0;
	while (fgets(buf,1023,fd))
	{
		char *copy;
	
		buf[strlen(buf)-1] = 0;

		copy = strdup(buf);

		listAddStringId(&extension->quotes,copy,copy);
	}

	fclose(fd);

	wonLog(wcp->ctx,wcp->modctx,"using quote file '%s'.  %lu quotes.",wcp->buffer->str[1],extension->quotes.items);

	wonMemoryFree(extension->quoteFile);

	extension->quoteFile	= strdup(wcp->buffer->str[1]);

	return WON_STATUS_SUCCESS;
}

uint32 ircPrivmsg(WON_IRC_PACKET *wip)
{
	if (wip->buffer->words < 4)
		return WON_STATUS_SUCCESS;

	if (!strcmp(wip->buffer->str[3],":.quote"))
		cmdQuote(wip);
	else if (!strcmp(wip->buffer->str[3],":.help"))
		cmdHelp(wip);
	else if (!strcmp(wip->buffer->str[3],":.get"))
		cmdGet(wip);
	else if (!strcmp(wip->buffer->str[3],":afrohug"))
		cmdAfroHug(wip);
	else if (!strcmp(wip->buffer->str[3],":.addquote"))
		cmdAddQuote(wip);
	else if (!strcmp(wip->buffer->str[3],":.search"))
		cmdSearch(wip);

	return WON_STATUS_SUCCESS;
}

void cmdQuote(WON_IRC_PACKET *wip)
{
	QUOTE_MOD_EXTENSION *extension = (QUOTE_MOD_EXTENSION *)wip->modctx->moduleExtension;

	if (!extension->quotes.items)
		return;

	printQuote(extension,
			wip->ctx,
			wip->irc,
			(wip->buffer->str[2][0] == '#')?wip->buffer->str[2]:wip->fromNick,
			(rand()%extension->quotes.items));
}

void cmdHelp(WON_IRC_PACKET *wip)
{
	wonIrcSend(wip->ctx,wip->irc,"PRIVMSG %s :[wonton: \002quote.so\002            RIP jefuzick   ]\n",wip->fromNick);
	wonIrcSend(wip->ctx,wip->irc,"PRIVMSG %s :[  cmds: .addquote .get .help .quote afrohug]\n",wip->fromNick);
	wonIrcSend(wip->ctx,wip->irc,"PRIVMSG %s :[        .search                            ]\n",wip->fromNick);
}

void cmdGet(WON_IRC_PACKET *wip)
{
	QUOTE_MOD_EXTENSION *extension = (QUOTE_MOD_EXTENSION *)wip->modctx->moduleExtension;

	if (wip->buffer->words < 5)
	{
		wonIrcSend(wip->ctx,wip->irc,"PRIVMSG %s :let's try this again, this time specify a number.\n",wip->fromNick);
		return;
	}

	printQuote(extension,
		wip->ctx,
		wip->irc,
		(wip->buffer->str[2][0] == '#')?wip->buffer->str[2]:wip->fromNick,
		strtoul(wip->buffer->str[4],NULL,10));
}

void cmdAfroHug(WON_IRC_PACKET *wip)
{
	wonIrcSend(wip->ctx,wip->irc,"PRIVMSG %s :\001ACTION huggles %s\001\n",wip->buffer->str[2],wip->fromNick);
}

void cmdAddQuote(WON_IRC_PACKET *wip)
{
	QUOTE_MOD_EXTENSION *extension = (QUOTE_MOD_EXTENSION *)wip->modctx->moduleExtension;
	LINKED_LIST_ITEM *item;
	char *buf = NULL;
	FILE *fd;

	if (wip->buffer->words < 5)
	{
		wonIrcSend(wip->ctx,wip->irc,"PRIVMSG %s :that's pretty fucking funny, quoting nothing again are you?\n",wip->fromNick);
		return;
	}

	buf = (char *)wonMemoryAllocate(1024,0);
	buf[1023] = 0;
	cacheCompileWords(wip->buffer,4,wip->buffer->words," ",buf,1023);

	listAddStringId(&extension->quotes,buf,buf);

	wonIrcSend(wip->ctx,wip->irc,"PRIVMSG %s :added.  %lu quotes exist.\n",wip->fromNick,extension->quotes.items);

	if (!(fd = fopen(extension->quoteFile,"w")))
		return;
	
	for (item = listGetFirstItem(&extension->quotes);
		item && item->value;
		item = listGetNextItem(item))
		fprintf(fd,"%s\n",(char *)item->value);

	fclose(fd);
}

void cmdSearch(WON_IRC_PACKET *wip)
{
	QUOTE_MOD_EXTENSION *extension = (QUOTE_MOD_EXTENSION *)wip->modctx->moduleExtension;
	LINKED_LIST_ITEM *item;
	char buf[1024];
	int x = 0, matches = 0;

	if (wip->buffer->words < 5)
	{
		wonIrcSend(wip->ctx,wip->irc,"PRIVMSG %s :at least supply something to search for fuckface.\n",wip->fromNick);
		return;
	}

	buf[1023] = 0;
	cacheCompileWords(wip->buffer,4,wip->buffer->words," ",buf,1023);
	
	wonIrcSend(wip->ctx,wip->irc,"PRIVMSG %s :searching for '%s'\nPRIVMSG %s :Matches: ",wip->fromNick,buf,wip->fromNick);

	for (item = listGetFirstItem(&extension->quotes);
		item && item->value;
		item = listGetNextItem(item), x++)
	{
		if (strstr((char *)item->value,buf))
		{
			wonIrcSend(wip->ctx,wip->irc,"%i ",x);
			matches++;
		}
	}

	wonIrcSend(wip->ctx,wip->irc,"\nPRIVMSG %s :Found \037%i\037 matches.\n",wip->fromNick,matches);
}

void printQuote(QUOTE_MOD_EXTENSION *extension, WON_CTX *ctx,WON_IRC_CONNECTION *conn,const char *target,unsigned long quoteNumber)
{
	LINKED_LIST_ITEM *item;
	unsigned long x = 0;

	for (item = listGetFirstItem(&extension->quotes);
		item && item->value && x < quoteNumber;
		item = listGetNextItem(item), x++);

	if (item && item->value)
	{
		wonIrcSend(ctx,conn,"PRIVMSG %s :[%lu/%lu]\n",target,x,extension->quotes.items);
		wonIrcSend(ctx,conn,"PRIVMSG %s :%s\n",target,(char *)item->value);
	}
	else
		wonIrcSend(ctx,conn,"PRIVMSG %s :No such quote, bitch.\n",target);
}
