#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <time.h>

#include "won.h"
#include "wonkrnl.h"

struct {
	const char 	*str;
	uint32 		bit;
} dcc_access_relation[] = {
	{ "OPS",		WONKRNL_ACCESS_OPS 		},
	{ "AUTO_OPS",		WONKRNL_ACCESS_AUTO_OPS 	},
	{ "MASTER",		WONKRNL_ACCESS_MASTER		},
	{ "INVITES",		WONKRNL_ACCESS_INVITES		}, 
	{ "KICKS",		WONKRNL_ACCESS_KICKS		},
	{ "BANS",		WONKRNL_ACCESS_BANS		},
	{ "MSGS",		WONKRNL_ACCESS_MSGS		},
	{ NULL,			0				},
};

uint32 dccCheckGlobalPermissions(WONKRNL_USER *user,uint32 level)
{
	return IS_SET(user->gaccess,level);
}

uint32 dccCheckChannelPermissions(WONKRNL_USER *user,const char *channel,uint32 level)
{
	WONKRNL_USER_CHANNEL *userChannel = wonkrnlUserChannelHas(user,channel);

	if (!userChannel)
		return 0;

	return IS_SET(userChannel->access,level);
}

void dccLinkUp(WON_MODULE_CTX *modctx,WON_DCC_CONNECTION *conn)
{
	wonLog(modctx->wonctx,modctx,"DCC Link up! (%hi.%hi.%hi.%hi)",PRINT_IP(conn->addr));
}

void dccLinkDown(WON_MODULE_CTX *modctx,WON_DCC_CONNECTION *conn)
{
	WONKRNL_DCC_EXTENSION *dccExtension = (WONKRNL_DCC_EXTENSION *)conn->dccExtension;

	wonLog(modctx->wonctx,modctx,"DCC Link down!");

	if (dccExtension && dccExtension->user)
		wonDccSendAll(modctx->wonctx,"%s %s left me for another woman.\n",WONKRNL_DCC_STR,dccExtension->user->username);

	wonMemoryFree(conn->dccExtension);
}

void dccEstablishConnection(WON_IRC_PACKET *wip)
{
	WONKRNL_DCC_EXTENSION *ext;
	WON_DCC_CONNECTION *dcc;
	unsigned long resolveaddr;
	WONKRNL_USER *user;
	uint32 port;
	int fd;

	if (wip->buffer->words < 8)
		return;

	user = configLookupUserByMask(wip->irc,wip->buffer->str[0]);

	if (!user)
	{
		wonLog(wip->ctx,wip->modctx,"dcc: %s attempted to dcc chat me but was denied.",wip->fromNick);
		return;
	}

	port = atoi(wip->buffer->str[7]);

	if (port < 1024)
	{
		wonLog(wip->ctx,wip->modctx,"dcc: chat attempt from %s with invalid port.",wip->fromNick);
		return;
	}

	fd = wonCreateTcpConnection(0,wip->buffer->str[6],port,&resolveaddr);

	dcc = (WON_DCC_CONNECTION *)wonMemoryAllocate(sizeof(WON_DCC_CONNECTION),1);
	ext = (WONKRNL_DCC_EXTENSION *)wonMemoryAllocate(sizeof(WONKRNL_DCC_EXTENSION),1);

	ext->network 	  = (WONKRNL_NETWORK *)wip->irc->ircExtension;
	ext->irc	  = wip->irc;
	ext->echo	  = 1;
	ext->loginTime	  = time(NULL);
	ext->idleTime     = time(NULL);
	strncpy(ext->currentNickname,wip->fromNick,WON_MAX_NICKNAME_SIZE-1);

	dcc->fd		  = fd;
	dcc->addr	  = resolveaddr;
	dcc->port	  = port;
	dcc->dccExtension = (void *)ext;

	wonDccRegisterConnection(wip->ctx,dcc);

	wonDccSend(wip->ctx,dcc,"%s\n",ext->network->dccUserPrompt);
}

uint32 dccLowlevel(WON_MODULE_CTX *modctx,WON_PACKET *packet)
{
	WONKRNL_DCC_EXTENSION *dccExtension;

	if (packet->type != WON_CONNECTION_DCC)	/* Pass through. */
		return WON_STATUS_SUCCESS;

	if (packet->pkt.dcc->buffer->words == 0)
		return WON_STATUS_FAILURE;
	
	dccExtension = (WONKRNL_DCC_EXTENSION *)packet->pkt.dcc->dcc->dccExtension;

	if (dccExtension && dccExtension->user)
		dccExtension->idleTime = time(NULL);

	if (IS_SET(packet->pkt.dcc->dcc->state,WON_DCC_STATE_AUTH) && (packet->pkt.dcc->buffer->str[0][0] == '.'))
	{
		WON_DCC_CONNECTION *currDcc;
		WONKRNL_DCC_EXTENSION *sendExtension;
		int x = 0;

		if (!strcmp(packet->pkt.dcc->buffer->str[0],".passwd"))
			return WON_STATUS_SUCCESS;

		while (wonDccEnum(packet->pkt.dcc->ctx,x++,&currDcc) != WON_STATUS_FAILURE)
		{
			sendExtension = (WONKRNL_DCC_EXTENSION *)currDcc->dccExtension;

			if (sendExtension && sendExtension->user)
			{
				if (IS_SET(sendExtension->user->gaccess,WONKRNL_ACCESS_MASTER) && 
					((currDcc != packet->pkt.dcc->dcc) && 
						(dccExtension->echo)))
				{
					wonDccSend(packet->pkt.dcc->ctx,currDcc,
								"#%s# %s\n",
								dccExtension->user->username,
								packet->pkt.dcc->buffer->buf+1);
				}
			}
		}

		return WON_STATUS_SUCCESS;
	}

	if (!IS_SET(packet->pkt.dcc->dcc->state,WON_DCC_STATE_USERNAME))
	{
		WONKRNL_USER *user;
	
		user = (WONKRNL_USER *)listGetValueStringId(&dccExtension->network->users,packet->pkt.dcc->buffer->str[0]);

		if (!user)
		{
			wonDccDeregisterConnection(packet->pkt.dcc->ctx,packet->pkt.dcc->dcc);
			return WON_STATUS_FAILURE;
		}

		dccExtension->user = user;

		wonDccSend(packet->pkt.dcc->ctx,packet->pkt.dcc->dcc,"%s\n",dccExtension->network->dccPassPrompt);

		BIT_SET(packet->pkt.dcc->dcc->state,WON_DCC_STATE_USERNAME);
	}
	else if ((dccExtension->user) && (packet->pkt.dcc->buffer->words) && (!IS_SET(packet->pkt.dcc->dcc->state,WON_DCC_STATE_AUTH)))
	{
		WONKRNL_DCC_EXTENSION *dccExtension = (WONKRNL_DCC_EXTENSION *)packet->pkt.dcc->dcc->dccExtension;
		unsigned char md5val[16];
		int x = 0;

		md5Calc(md5val,packet->pkt.dcc->buffer->str[0],strlen(packet->pkt.dcc->buffer->str[0]));

		for (x = 0; x < 16; x++)
		{
			if (md5val[x] != dccExtension->user->password[x])
			{
				wonDccDeregisterConnection(packet->pkt.dcc->ctx,packet->pkt.dcc->dcc);
				return WON_STATUS_FAILURE;
			}
		}

		dccSendMotd(packet->pkt.dcc->ctx,packet->pkt.dcc->dcc,dccExtension);	
		
		BIT_SET(packet->pkt.dcc->dcc->state,WON_DCC_STATE_AUTH);

		wonDccSendAll(packet->pkt.dcc->ctx,"%s %s has become one with wonton.\n",WONKRNL_DCC_STR,dccExtension->user->username);
	}
	else
	{
		/* this goes out to the party line */
		WONKRNL_DCC_EXTENSION *dccExtension;
		WON_DCC_CONNECTION *currDcc;
		char buf[512];
		int x = 0;

		buf[511] = 0;
		cacheCompileWords(packet->pkt.dcc->buffer,0,packet->pkt.dcc->buffer->words," ",buf,511);

		while (wonDccEnum(packet->pkt.dcc->ctx,x++,&currDcc) != WON_STATUS_FAILURE)
		{
			dccExtension = (WONKRNL_DCC_EXTENSION *)packet->pkt.dcc->dcc->dccExtension;

			if ((currDcc == packet->pkt.dcc->dcc) && (!dccExtension->echo))
				continue;
	
			wonDccSend(packet->pkt.dcc->ctx,currDcc,
							"[%s] %s\n",
							dccExtension->user->username,
							buf);
		}
	}

	return WON_STATUS_FAILURE;
}

void dccSendMotd(WON_CTX *ctx,WON_DCC_CONNECTION *dcc, WONKRNL_DCC_EXTENSION *dccExtension)
{
	FILE *fd = (dccExtension->network->motdFile)?fopen(dccExtension->network->motdFile,"r"):NULL;
	char buffer[512];

	if (!fd)
	{
		wonDccSend(ctx,dcc,"welcome to the wonderful world of wonton.\n");
		return;
	}

	buffer[511] = 0;
	while (fgets(buffer,511,fd))
		wonDccSend(ctx,dcc,"%s",buffer);

	fclose(fd);
}

uint32 dccAddUser(WON_DCC_PACKET *wdp)
{
	WONKRNL_DCC_EXTENSION *dccExtension = (WONKRNL_DCC_EXTENSION *)wdp->dcc->dccExtension;

	if (wdp->buffer->words < 2)
	{
		wonDccSend(wdp->ctx,wdp->dcc,"Usage: \037adduser\037 [\002username\002]\n");
		wonDccSend(wdp->ctx,wdp->dcc,"      Adds a user.\n");
		return WON_STATUS_SUCCESS;
	}

	if (!dccCheckGlobalPermissions(dccExtension->user,WONKRNL_ACCESS_MASTER))
		return WON_STATUS_SUCCESS;

	if (wonkrnlUserAdd(dccExtension->network,wdp->buffer->str[1]))
		wonDccSend(wdp->ctx,wdp->dcc,"Successfully created user \002%s\002.\n",wdp->buffer->str[1]);
	else
		wonDccSend(wdp->ctx,wdp->dcc,"Failed to create user \002%s\002.\n",wdp->buffer->str[1]);

	return WON_STATUS_SUCCESS;
}

uint32 dccDelUser(WON_DCC_PACKET *wdp)
{
	WONKRNL_DCC_EXTENSION *dccExtension = (WONKRNL_DCC_EXTENSION *)wdp->dcc->dccExtension;
	char usernameCopy[WON_MAX_NICKNAME_SIZE];
	WON_STATUS retval;

	if (wdp->buffer->words < 2)
	{
		wonDccSend(wdp->ctx,wdp->dcc,"Usage: \037deluser\037 [\002username\002]\n");
		wonDccSend(wdp->ctx,wdp->dcc,"      Deletes a user.\n");
		return WON_STATUS_SUCCESS;
	}

	if (!dccCheckGlobalPermissions(dccExtension->user,WONKRNL_ACCESS_MASTER))
		return WON_STATUS_SUCCESS;

	strncpy(usernameCopy,wdp->buffer->str[1],WON_MAX_NICKNAME_SIZE-1);
	
	retval = dccBoot(wdp);		/* Kick the user off first. */

	if (wonkrnlUserDelete(dccExtension->network,usernameCopy) == WON_STATUS_SUCCESS)
		wonDccSend(wdp->ctx,wdp->dcc,"Successfully deleted user \002%s\002.\n",usernameCopy);
	else
		wonDccSend(wdp->ctx,wdp->dcc,"Failed to delete user \002%s\002.\n",usernameCopy);

	return retval;
}

uint32 dccBoot(WON_DCC_PACKET *wdp)
{
	WONKRNL_DCC_EXTENSION *dccExtension = (WONKRNL_DCC_EXTENSION *)wdp->dcc->dccExtension;
	WON_DCC_CONNECTION *dcc;
	LINKED_LIST_ITEM *item;
	WON_STATUS retval = WON_STATUS_SUCCESS;

	if (wdp->buffer->words < 2)
	{
		wonDccSend(wdp->ctx,wdp->dcc,"Usage: \037boot\037 [\002username\002]\n");
		wonDccSend(wdp->ctx,wdp->dcc,"      Boots a user from dcc.\n");
		return WON_STATUS_SUCCESS;
	}

	if (!dccCheckGlobalPermissions(dccExtension->user,WONKRNL_ACCESS_MASTER))
		return WON_STATUS_SUCCESS;

	item = listGetFirstItem(&wdp->ctx->dccConnections);

	while (item && item->value)
	{
		dcc = (WON_DCC_CONNECTION *)item->value;
		dccExtension = (WONKRNL_DCC_EXTENSION *)dcc->dccExtension;
	
		if ((dccExtension->user) && (!strcmp(dccExtension->user->username,wdp->buffer->str[1])))
		{
			if (dcc == wdp->dcc)
				retval = WON_STATUS_CONNCLOSED;

			wonDccSend(wdp->ctx,dcc,"Your visa on existance has expired.\n");

			wonDccDeregisterConnection(wdp->ctx,dcc);
			item = listGetFirstItem(&wdp->ctx->dccConnections);

			wonDccSend(wdp->ctx,wdp->dcc,"Successfully booted user \002%s\002.\n",wdp->buffer->str[1]);
			continue;
		}

		item = listGetNextItem(item);
	}

	return retval;
}

uint32 dccUserInfo(WON_DCC_PACKET *wdp)
{
	WONKRNL_DCC_EXTENSION *dccExtension = (WONKRNL_DCC_EXTENSION *)wdp->dcc->dccExtension;
	WONKRNL_USER_CHANNEL *userChannel;
	LINKED_LIST_ITEM *item;
	WONKRNL_USER *user;
	int x = 0;

	if (wdp->buffer->words < 2)
	{
		wonDccSend(wdp->ctx,wdp->dcc,"Usage: \037userinfo\037 [\002username\002]\n");
		wonDccSend(wdp->ctx,wdp->dcc,"      Returns information about a specific user.\n");
		return WON_STATUS_SUCCESS;
	}

	user = wonkrnlUserHas(dccExtension->network,wdp->buffer->str[1]);

	if (!user)
	{
		wonDccSend(wdp->ctx,wdp->dcc,"User \002%s\002 does not exist.\n",wdp->buffer->str[1]);
		return WON_STATUS_SUCCESS;
	}

	wonDccSend(wdp->ctx,wdp->dcc,"Information for user [\037%s\037]\n",wdp->buffer->str[1]);
	wonDccSend(wdp->ctx,wdp->dcc,"  Global Access\037:\037 ");

	for (x = 0; dcc_access_relation[x].str; x++)
	{
		if (IS_SET(user->gaccess,dcc_access_relation[x].bit))
			wonDccSend(wdp->ctx,wdp->dcc,"%s ",dcc_access_relation[x].str);
	}

	wonDccSend(wdp->ctx,wdp->dcc,"\n  Channels (%02i)\037:\037 ",user->channels.items);

	for (item = listGetFirstItem(&user->channels);
		item && item->value;
		item = listGetNextItem(item))
	{
		userChannel = (WONKRNL_USER_CHANNEL *)item->value;

		wonDccSend(wdp->ctx,wdp->dcc,"%s ",userChannel->channel->channel);
	}
	
	wonDccSend(wdp->ctx,wdp->dcc,"\n  Hosts    (%02i)\037:\037 ",user->hosts.items);

	for (item = listGetFirstItem(&user->hosts);
		item && item->value;
		item = listGetNextItem(item))
		wonDccSend(wdp->ctx,wdp->dcc,"%s ",(char *)item->value);

	wonDccSend(wdp->ctx,wdp->dcc,"\n");
	
	return WON_STATUS_SUCCESS;
}

uint32 dccUattr(WON_DCC_PACKET *wdp)
{
	WONKRNL_DCC_EXTENSION *dccExtension = (WONKRNL_DCC_EXTENSION *)wdp->dcc->dccExtension;
	WONKRNL_USER *user;
	int x = 0, y = 0;

	if (wdp->buffer->words < 3)
	{
		wonDccSend(wdp->ctx,wdp->dcc,"Usage: \037uattr\037 [\002username\002] [\002+FLAG1 -FLAG2 ...\002]\n");
		wonDccSend(wdp->ctx,wdp->dcc,"      Sets a users global access flags.\n");
		return WON_STATUS_SUCCESS;
	}

	if (!dccCheckGlobalPermissions(dccExtension->user,WONKRNL_ACCESS_MASTER))
		return WON_STATUS_SUCCESS;

	user = wonkrnlUserHas(dccExtension->network,wdp->buffer->str[1]);

	if (!user)
	{
		wonDccSend(wdp->ctx,wdp->dcc,"User \002%s\002 does not exist.\n",wdp->buffer->str[1]);
		return WON_STATUS_SUCCESS;
	}

	for (x = 2; x < wdp->buffer->words && wdp->buffer->str[x]; x++)
	{
		int allMinus = 0;

		if (!strcmp(wdp->buffer->str[x]+1,"ALL"))
		{
			if (wdp->buffer->str[x][0] == '-')
				allMinus = 1;
			else
				allMinus = 2;
		}
				
		for (y =0; dcc_access_relation[y].str; y++)
		{
			if (!strcmp(dcc_access_relation[y].str,wdp->buffer->str[x]+1))
			{
				if ((wdp->buffer->str[x][0] == '+') || (allMinus == 2))
					BIT_SET(user->gaccess,dcc_access_relation[y].bit);
				else if ((wdp->buffer->str[x][0] == '-') || (allMinus == 1))
					BIT_REMOVE(user->gaccess,dcc_access_relation[y].bit);
			}
		}
	}

	wonDccSend(wdp->ctx,wdp->dcc,"Successfully changed global access for \002%s\002.\n",wdp->buffer->str[1]);
	
	return WON_STATUS_SUCCESS;
}

uint32 dccCattr(WON_DCC_PACKET *wdp)
{
	WONKRNL_DCC_EXTENSION *dccExtension = (WONKRNL_DCC_EXTENSION *)wdp->dcc->dccExtension;
	WONKRNL_USER *user;
	int x, y;

	if (wdp->buffer->words < 4)
	{
		wonDccSend(wdp->ctx,wdp->dcc,"Usage: \037uattr\037 [\002username\002] [\002channel/*\002] [\002+FLAG1 -FLAG2 ...\002]\n");
		wonDccSend(wdp->ctx,wdp->dcc,"      Sets a users channel access flags.\n");
		return WON_STATUS_SUCCESS;
	}

	if ((!dccCheckGlobalPermissions(dccExtension->user,WONKRNL_ACCESS_MASTER)) &&
		(!dccCheckChannelPermissions(dccExtension->user,wdp->buffer->str[2],WONKRNL_ACCESS_MASTER)))
		return WON_STATUS_SUCCESS;

	user = wonkrnlUserHas(dccExtension->network,wdp->buffer->str[1]);

	if (!strcmp(wdp->buffer->str[2],"*"))
	{
		WONKRNL_CHANNEL *channel;
		LINKED_LIST_ITEM *item;

		for (item = listGetFirstItem(&dccExtension->network->channels);
				item && item->value;
				item = listGetNextItem(item))
		{
			WONKRNL_USER_CHANNEL *userChannel;
			channel = (WONKRNL_CHANNEL *)item->value;

			userChannel = wonkrnlUserChannelHas(user,channel->channel);
			
			if (!userChannel)
				userChannel = wonkrnlUserChannelAdd(user,channel->channel);		/* Add if it doesn't exist. */

			for (x = 3; x < wdp->buffer->words && wdp->buffer->str[x]; x++)
			{
				int allMinus = 0;

				if (!strcmp(wdp->buffer->str[x]+1,"ALL"))
				{
					if (wdp->buffer->str[x][0] == '-')
						allMinus = 1;
					else
						allMinus = 2;
				}
					
				for (y =0; dcc_access_relation[y].str; y++)
				{
					if (!strcmp(dcc_access_relation[y].str,wdp->buffer->str[x]+1))
					{
						if ((wdp->buffer->str[x][0] == '+') || (allMinus == 2))
							BIT_SET(userChannel->access,dcc_access_relation[y].bit);
						else if ((wdp->buffer->str[x][0] == '-') || (allMinus == 1))
							BIT_REMOVE(userChannel->access,dcc_access_relation[y].bit);
					}
				}
			}
		
			wonDccSend(wdp->ctx,wdp->dcc,"Successfully changed flags for user \002%s\002 on channel \002%s\002.\n",wdp->buffer->str[1],channel->channel);
		}
	}
	else
	{
		WONKRNL_CHANNEL *channel = wonkrnlChannelHas(dccExtension->network,wdp->buffer->str[2]);
		WONKRNL_USER_CHANNEL *userChannel;
		
		if (!channel)
		{
			wonDccSend(wdp->ctx,wdp->dcc,"The channel \002%s\002 does not exist.\n",wdp->buffer->str[2]);
			return WON_STATUS_SUCCESS;
		}

		userChannel = wonkrnlUserChannelHas(user,channel->channel);
		
		if (!userChannel)
			userChannel = wonkrnlUserChannelAdd(user,channel->channel);		/* Add if it doesn't exist. */

		for (x = 3; x < wdp->buffer->words && wdp->buffer->str[x]; x++)
		{
			int allMinus = 0;

			if (!strcmp(wdp->buffer->str[x]+1,"ALL"))
			{
				if (wdp->buffer->str[x][0] == '-')
					allMinus = 1;
				else
					allMinus = 2;
			}

			for (y =0; dcc_access_relation[y].str; y++)
			{
				if (!strcmp(dcc_access_relation[y].str,wdp->buffer->str[x]+1))
				{
					if ((wdp->buffer->str[x][0] == '+') || (allMinus == 2))
						BIT_SET(userChannel->access,dcc_access_relation[y].bit);
					else if ((wdp->buffer->str[x][0] == '-') || (allMinus == 1))
						BIT_REMOVE(userChannel->access,dcc_access_relation[y].bit);
				}
			}
		}
	
		wonDccSend(wdp->ctx,wdp->dcc,"Successfully changed flags for user \002%s\002 on channel \002%s\002.\n",wdp->buffer->str[1],wdp->buffer->str[2]);
	}

	
	return WON_STATUS_SUCCESS;
}

uint32 dccAddHost(WON_DCC_PACKET *wdp)
{
	WONKRNL_DCC_EXTENSION *dccExtension = (WONKRNL_DCC_EXTENSION *)wdp->dcc->dccExtension;
	WONKRNL_USER *user;
	const char *newHost;

	if (wdp->buffer->words < 3)
	{
		wonDccSend(wdp->ctx,wdp->dcc,"Usage: \037addhost\037 [\002username\002] [\002nick!ident@host.com\002]\n");
		wonDccSend(wdp->ctx,wdp->dcc,"      Adds a permitted host for a user.\n");
		return WON_STATUS_SUCCESS;
	}

	if ((strcmp(dccExtension->user->username,wdp->buffer->str[1])) &&
		(!dccCheckGlobalPermissions(dccExtension->user,WONKRNL_ACCESS_MASTER)))
		return WON_STATUS_SUCCESS;

	user = wonkrnlUserHas(dccExtension->network,wdp->buffer->str[1]);

	if (!user)
	{
		wonDccSend(wdp->ctx,wdp->dcc,"User \002%s\002 does not exist.\n",wdp->buffer->str[1]);
		return WON_STATUS_SUCCESS;
	}

	newHost = wonkrnlUserHostAdd(user,wdp->buffer->str[2]);

	if (!newHost)
	{
		wonDccSend(wdp->ctx,wdp->dcc,"A more explicit host already exists for user \002%s\002.\n",wdp->buffer->str[1]);
		return WON_STATUS_SUCCESS;
	}

	wonDccSend(wdp->ctx,wdp->dcc,"Successfully added host \037%s\037 for user \002%s\002.\n",wdp->buffer->str[2],wdp->buffer->str[1]);

	return WON_STATUS_SUCCESS;
}

uint32 dccDelHost(WON_DCC_PACKET *wdp)
{
	WONKRNL_DCC_EXTENSION *dccExtension = (WONKRNL_DCC_EXTENSION *)wdp->dcc->dccExtension;
	WONKRNL_USER *user;

	if (wdp->buffer->words < 3)
	{
		wonDccSend(wdp->ctx,wdp->dcc,"Usage: \037delhost\037 [\002username\002] [\002nick!ident@host.com\002]\n");
		wonDccSend(wdp->ctx,wdp->dcc,"      Deletes any hosts that match the specified mask for a user.\n");
		return WON_STATUS_SUCCESS;
	}

	if ((strcmp(dccExtension->user->username,wdp->buffer->str[1])) &&
		(!dccCheckGlobalPermissions(dccExtension->user,WONKRNL_ACCESS_MASTER)))
		return WON_STATUS_SUCCESS;

	user = wonkrnlUserHas(dccExtension->network,wdp->buffer->str[1]);

	if (!user)
	{
		wonDccSend(wdp->ctx,wdp->dcc,"User \002%s\002 does not exist.\n",wdp->buffer->str[1]);
		return WON_STATUS_SUCCESS;
	}

	if (wonkrnlUserHostDelete(user,wdp->buffer->str[2]) != WON_STATUS_SUCCESS)
	{
		wonDccSend(wdp->ctx,wdp->dcc,"No hosts matched mask \037%s\037 for user \002%s\002.\n",wdp->buffer->str[2],wdp->buffer->str[1]);
		return WON_STATUS_SUCCESS;
	}

	wonDccSend(wdp->ctx,wdp->dcc,"Successfully deleted host \037%s\037 for user \002%s\002.\n",wdp->buffer->str[2],wdp->buffer->str[1]);

	return WON_STATUS_SUCCESS;
}

uint32 dccPasswd(WON_DCC_PACKET *wdp)
{
	WONKRNL_DCC_EXTENSION *dccExtension = (WONKRNL_DCC_EXTENSION *)wdp->dcc->dccExtension;
	WONKRNL_USER *user;

	if (wdp->buffer->words < 3)
	{
		wonDccSend(wdp->ctx,wdp->dcc,"Usage: \037passwd\037 [\002username\002] [\002newpassword\002]\n");
		wonDccSend(wdp->ctx,wdp->dcc,"      Changes a users password.\n");
		return WON_STATUS_SUCCESS;
	}

	if ((strcmp(dccExtension->user->username,wdp->buffer->str[1])) &&
		(!dccCheckGlobalPermissions(dccExtension->user,WONKRNL_ACCESS_MASTER)))
		return WON_STATUS_SUCCESS;

	user = wonkrnlUserHas(dccExtension->network,wdp->buffer->str[1]);

	if (!user)
	{
		wonDccSend(wdp->ctx,wdp->dcc,"User \002%s\002 does not exist.\n",wdp->buffer->str[1]);
		return WON_STATUS_SUCCESS;
	}

	md5Calc(user->password,wdp->buffer->str[2],strlen(wdp->buffer->str[2]));

	wonDccSend(wdp->ctx,wdp->dcc,"Successfully changed password for user \002%s\002.\n",wdp->buffer->str[1]);

	return WON_STATUS_SUCCESS;
}

uint32 dccAddChan(WON_DCC_PACKET *wdp)
{
	WONKRNL_DCC_EXTENSION *dccExtension = (WONKRNL_DCC_EXTENSION *)wdp->dcc->dccExtension;

	if (wdp->buffer->words < 2)
	{
		wonDccSend(wdp->ctx,wdp->dcc,"Usage: \037addchan\037 [\002channel\002]\n");
		wonDccSend(wdp->ctx,wdp->dcc,"      Add a new channel.\n");
		return WON_STATUS_SUCCESS;
	}

	if (!dccCheckGlobalPermissions(dccExtension->user,WONKRNL_ACCESS_MASTER))
		return WON_STATUS_SUCCESS;

	if (!wonkrnlChannelAdd(dccExtension->network,wdp->buffer->str[1]))
		wonDccSend(wdp->ctx,wdp->dcc,"The channel \002%s\002 already exists.\n",wdp->buffer->str[1]);
	else	
		wonDccSend(wdp->ctx,wdp->dcc,"Successfully added channel \002%s\002.\n",wdp->buffer->str[1]);

	return WON_STATUS_SUCCESS;
}

uint32 dccDelChan(WON_DCC_PACKET *wdp)
{
	WONKRNL_DCC_EXTENSION *dccExtension = (WONKRNL_DCC_EXTENSION *)wdp->dcc->dccExtension;
	WONKRNL_CHANNEL *channel;

	if (wdp->buffer->words < 2)
	{
		wonDccSend(wdp->ctx,wdp->dcc,"Usage: \037delchan\037 [\002channel\002]\n");
		wonDccSend(wdp->ctx,wdp->dcc,"      Delete a channel.\n");
		return WON_STATUS_SUCCESS;
	}

	if (!dccCheckGlobalPermissions(dccExtension->user,WONKRNL_ACCESS_MASTER))
		return WON_STATUS_SUCCESS;

	channel = wonkrnlChannelHas(dccExtension->network,wdp->buffer->str[1]);		

	if (wonkrnlChannelDelete(dccExtension->network,wdp->buffer->str[1]) != WON_STATUS_SUCCESS)
		wonDccSend(wdp->ctx,wdp->dcc,"The channel \002%s\002 does not exist.\n",wdp->buffer->str[1]);
	else	
	{
		WON_DCC_CONNECTION *currDcc;
		int x = 0;

		while (wonDccEnum(wdp->ctx,x++,&currDcc) != WON_STATUS_FAILURE)
		{
			dccExtension = (WONKRNL_DCC_EXTENSION *)currDcc->dccExtension;

			if (dccExtension->currentChannel == channel)
				dccExtension->currentChannel = NULL;
		}

		wonDccSend(wdp->ctx,wdp->dcc,"Successfully deleted channel \002%s\002.\n",wdp->buffer->str[1]);
	}

	return WON_STATUS_SUCCESS;
}

uint32 dccChanInfo(WON_DCC_PACKET *wdp)
{
	if (wdp->buffer->words < 2)
	{
		wonDccSend(wdp->ctx,wdp->dcc,"Usage: \037chaninfo\037 [\002channel\002]\n");
		wonDccSend(wdp->ctx,wdp->dcc,"      Return information regarding a channel.\n");
		return WON_STATUS_SUCCESS;
	}

	return WON_STATUS_SUCCESS;
}

uint32 dccConsole(WON_DCC_PACKET *wdp)
{
	WONKRNL_DCC_EXTENSION *dccExtension = (WONKRNL_DCC_EXTENSION *)wdp->dcc->dccExtension;
	WONKRNL_USER_CHANNEL *userChannel;

	if (wdp->buffer->words < 2)
	{
		wonDccSend(wdp->ctx,wdp->dcc,"Usage: \037console\037 [\002channel\002]\n");
		wonDccSend(wdp->ctx,wdp->dcc,"      Change your current channel console.\n");
		return WON_STATUS_SUCCESS;
	}

	userChannel = wonkrnlUserChannelHas(dccExtension->user,wdp->buffer->str[1]);

	if (!userChannel)
	{
		wonDccSend(wdp->ctx,wdp->dcc,"The channel \002%s\002 does not exist or you do not have access.\n",wdp->buffer->str[1]);
		return WON_STATUS_SUCCESS;
	}

	dccExtension->currentChannel = userChannel->channel;

	wonDccSend(wdp->ctx,wdp->dcc,"Your active channel is now \002%s\002.\n",wdp->buffer->str[1]);

	return WON_STATUS_SUCCESS;
}

uint32 dccSay(WON_DCC_PACKET *wdp)
{
	WONKRNL_DCC_EXTENSION *dccExtension = (WONKRNL_DCC_EXTENSION *)wdp->dcc->dccExtension;
	char str[512];

	if (wdp->buffer->words < 2)
	{
		wonDccSend(wdp->ctx,wdp->dcc,"Usage: \037say\037 [\002text\002]\n");
		wonDccSend(wdp->ctx,wdp->dcc,"      Send a message of \002text\002 to your active channel.\n");
		return WON_STATUS_SUCCESS;
	}

	if (!dccExtension->currentChannel)
	{
		wonDccSend(wdp->ctx,wdp->dcc,"You do not have an active channel set.\n");
		return WON_STATUS_SUCCESS;
	}

	if (!IS_SET(dccExtension->currentChannel->state,WONKRNL_STATE_CHANNEL_IN))
	{
		wonDccSend(wdp->ctx,wdp->dcc,"I'm not currently in \002%s\002.\n",dccExtension->currentChannel->channel);
		return WON_STATUS_SUCCESS;
	}

	if (!dccExtension->irc)
	{
		wonDccSend(wdp->ctx,wdp->dcc,"I'm not currently on IRC.\n");
		return WON_STATUS_SUCCESS;
	}

	str[511] = 0;
	cacheCompileWords(wdp->buffer,1,wdp->buffer->words," ",str,511);

	wonIrcSend(wdp->ctx,dccExtension->irc,
					"PRIVMSG %s :%s\n",
					dccExtension->currentChannel->channel,
					str);

	wonDccSend(wdp->ctx,wdp->dcc,"\037%s\037> %s\n",dccExtension->currentChannel->channel,str);

	return WON_STATUS_SUCCESS;
}

uint32 dccOp(WON_DCC_PACKET *wdp)
{
	WONKRNL_DCC_EXTENSION *dccExtension = (WONKRNL_DCC_EXTENSION *)wdp->dcc->dccExtension;

	if (!dccExtension->currentChannel)
	{
		wonDccSend(wdp->ctx,wdp->dcc,"You must have an active channel set.\n");
		return WON_STATUS_SUCCESS;
	}

	if ((!dccCheckGlobalPermissions(dccExtension->user,WONKRNL_ACCESS_OPS)) && 
		(!dccCheckChannelPermissions(dccExtension->user,dccExtension->currentChannel->channel,WONKRNL_ACCESS_OPS)))
	{
		wonDccSend(wdp->ctx,wdp->dcc,"No.\n");
		return WON_STATUS_SUCCESS;
	}

	if (!IS_SET(dccExtension->currentChannel->state,WONKRNL_STATE_CHANNEL_IN))
	{
		wonDccSend(wdp->ctx,wdp->dcc,"I'm not currently in \002%s\002.\n",dccExtension->currentChannel->channel);
		return WON_STATUS_SUCCESS;
	}
	
	if (!IS_SET(dccExtension->currentChannel->state,WONKRNL_STATE_CHANNEL_OP))
	{
		wonDccSend(wdp->ctx,wdp->dcc,"I'm not currently opped in \002%s\002.\n",dccExtension->currentChannel->channel);
		return WON_STATUS_SUCCESS;
	}

	if (!dccExtension->irc)
	{
		wonDccSend(wdp->ctx,wdp->dcc,"I'm not currently on IRC.\n");
		return WON_STATUS_SUCCESS;
	}

	wonIrcSend(wdp->ctx,dccExtension->irc,
				"MODE %s +o %s\n",
				dccExtension->currentChannel->channel,
				dccExtension->currentNickname);

	return WON_STATUS_SUCCESS;
}

uint32 dccInvite(WON_DCC_PACKET *wdp)
{
	WONKRNL_DCC_EXTENSION *dccExtension = (WONKRNL_DCC_EXTENSION *)wdp->dcc->dccExtension;

	if (!dccExtension->currentChannel)
	{
		wonDccSend(wdp->ctx,wdp->dcc,"You must have an active channel set.\n");
		return WON_STATUS_SUCCESS;
	}

	if ((!dccCheckGlobalPermissions(dccExtension->user,WONKRNL_ACCESS_INVITES)) && 
		(!dccCheckChannelPermissions(dccExtension->user,dccExtension->currentChannel->channel,WONKRNL_ACCESS_INVITES)))
	{
		wonDccSend(wdp->ctx,wdp->dcc,"No.\n");
		return WON_STATUS_SUCCESS;
	}

	if (!IS_SET(dccExtension->currentChannel->state,WONKRNL_STATE_CHANNEL_IN))
	{
		wonDccSend(wdp->ctx,wdp->dcc,"I'm not currently in \002%s\002.\n",dccExtension->currentChannel->channel);
		return WON_STATUS_SUCCESS;
	}

	if (!dccExtension->irc)
	{
		wonDccSend(wdp->ctx,wdp->dcc,"I'm not currently on IRC.\n");
		return WON_STATUS_SUCCESS;
	}

	wonIrcSend(wdp->ctx,dccExtension->irc,
				"INVITE %s %s\n",
				dccExtension->currentNickname,
				dccExtension->currentChannel->channel);

	wonDccSend(wdp->ctx,wdp->dcc,"Inviting you to \002%s\002.\n",dccExtension->currentChannel->channel);

	return WON_STATUS_SUCCESS;
}

uint32 dccKick(WON_DCC_PACKET *wdp)
{
	WONKRNL_DCC_EXTENSION *dccExtension = (WONKRNL_DCC_EXTENSION *)wdp->dcc->dccExtension;
	char reason[256];

	if (wdp->buffer->words < 2)
	{
		wonDccSend(wdp->ctx,wdp->dcc,"Usage: \037kick\037 [\002nick\002] [\002reason\002]\n");
		wonDccSend(wdp->ctx,wdp->dcc,"      Kick a specified nick from your active channel.\n");
		return WON_STATUS_SUCCESS;
	}
	
	if (!dccExtension->currentChannel)
	{
		wonDccSend(wdp->ctx,wdp->dcc,"You must have an active channel set.\n");
		return WON_STATUS_SUCCESS;
	}

	if ((!dccCheckGlobalPermissions(dccExtension->user,WONKRNL_ACCESS_KICKS)) && 
		(!dccCheckChannelPermissions(dccExtension->user,dccExtension->currentChannel->channel,WONKRNL_ACCESS_KICKS)))
	{
		wonDccSend(wdp->ctx,wdp->dcc,"No.\n");
		return WON_STATUS_SUCCESS;
	}

	if (!IS_SET(dccExtension->currentChannel->state,WONKRNL_STATE_CHANNEL_IN))
	{
		wonDccSend(wdp->ctx,wdp->dcc,"I'm not currently in \002%s\002.\n",dccExtension->currentChannel->channel);
		return WON_STATUS_SUCCESS;
	}
	
	if (!IS_SET(dccExtension->currentChannel->state,WONKRNL_STATE_CHANNEL_OP))
	{
		wonDccSend(wdp->ctx,wdp->dcc,"I'm not currently opped in \002%s\002.\n",dccExtension->currentChannel->channel);
		return WON_STATUS_SUCCESS;
	}

	if (!dccExtension->irc)
	{
		wonDccSend(wdp->ctx,wdp->dcc,"I'm not currently on IRC.\n");
		return WON_STATUS_SUCCESS;
	}

	reason[255] = 0;

	if (wdp->buffer->words > 2)
		cacheCompileWords(wdp->buffer,2,wdp->buffer->words," ",reason,255);
	else
		strncpy(reason,"in the end you die.",255);

	wonIrcSend(wdp->ctx,dccExtension->irc,
				"KICK %s %s :%s\n",
				dccExtension->currentChannel->channel,
				wdp->buffer->str[1],
				reason);

	wonDccSend(wdp->ctx,wdp->dcc,"[\037kick\037/\002%s\002] (%s)\n",
					wdp->buffer->str[1],
					reason);

	return WON_STATUS_SUCCESS;
}

uint32 dccAddBan(WON_DCC_PACKET *wdp)
{
	WONKRNL_DCC_EXTENSION *dccExtension = (WONKRNL_DCC_EXTENSION *)wdp->dcc->dccExtension;
	char reason[256];

	if (wdp->buffer->words < 3)
	{
		wonDccSend(wdp->ctx,wdp->dcc,"Usage: \037addban\037 [\002channel/*\002] [\002mask\002] [\002reason\002]\n");
		wonDccSend(wdp->ctx,wdp->dcc,"      Ban a specific mask from one or more channels.\n");
		return WON_STATUS_SUCCESS;
	}

	reason[255] = 0;
	if (wdp->buffer->words > 3)
		cacheCompileWords(wdp->buffer,3,wdp->buffer->words," ",reason,255);
	else
		strncpy(reason,"deprecate yourself.",255);
	
	if (!strcmp(wdp->buffer->str[1],"*"))
	{
		WONKRNL_USER_CHANNEL *channel;
		LINKED_LIST_ITEM *item;

		for (item = listGetFirstItem(&dccExtension->user->channels);
				item && item->value;
				item = listGetNextItem(item))
		{
			channel = (WONKRNL_USER_CHANNEL *)item->value;

			if ((!dccCheckGlobalPermissions(dccExtension->user,WONKRNL_ACCESS_BANS)) 
				&& (!dccCheckChannelPermissions(dccExtension->user,channel->channel->channel,WONKRNL_ACCESS_BANS)))
			{
				wonDccSend(wdp->ctx,wdp->dcc,"No permissions for bans on \002%s\002.\n",channel->channel->channel);
				return WON_STATUS_SUCCESS;
			}

			if (wonkrnlBanAdd(channel->channel,wdp->buffer->str[2],reason))
			{
				wonIrcSend(wdp->ctx,dccExtension->irc,
						"MODE %s +b %s\n",
						channel->channel->channel,
						wdp->buffer->str[2]);
				wonDccSend(wdp->ctx,wdp->dcc,"Successfully added ban to channel \002%s\002.\n",channel->channel->channel);
			}
			else
				wonDccSend(wdp->ctx,wdp->dcc,"A more specific ban already exists on \002%s\002.\n",channel->channel->channel);
		}
	}
	else
	{
		WONKRNL_USER_CHANNEL *channel = wonkrnlUserChannelHas(dccExtension->user,wdp->buffer->str[1]);

		if (!channel)
		{
			wonDccSend(wdp->ctx,wdp->dcc,"The channel \002%s\002 does not exist.\n",wdp->buffer->str[1]);
			return WON_STATUS_SUCCESS;
		}

		if ((!dccCheckGlobalPermissions(dccExtension->user,WONKRNL_ACCESS_BANS)) 
			&& (!dccCheckChannelPermissions(dccExtension->user,channel->channel->channel,WONKRNL_ACCESS_BANS)))
		{
			wonDccSend(wdp->ctx,wdp->dcc,"No.\n");
			return WON_STATUS_SUCCESS;
		}

		if (wonkrnlBanAdd(channel->channel,wdp->buffer->str[2],reason))
		{
			wonIrcSend(wdp->ctx,dccExtension->irc,
						"MODE %s +b %s\n",
						channel->channel->channel,
						wdp->buffer->str[2]);
			wonDccSend(wdp->ctx,wdp->dcc,"Successfully added ban to channel \002%s\002.\n",channel->channel->channel);
		}
		else
			wonDccSend(wdp->ctx,wdp->dcc,"A more specific ban already exists on \002%s\002.\n",channel->channel);
	}

	return WON_STATUS_SUCCESS;
}

uint32 dccDelBan(WON_DCC_PACKET *wdp)
{
	WONKRNL_DCC_EXTENSION *dccExtension = (WONKRNL_DCC_EXTENSION *)wdp->dcc->dccExtension;

	if (wdp->buffer->words < 3)
	{
		wonDccSend(wdp->ctx,wdp->dcc,"Usage: \037delban\037 [\002channel/*\002] [\002mask\002]\n");
		wonDccSend(wdp->ctx,wdp->dcc,"      Unban a specific mask from one or more channels.\n");
		return WON_STATUS_SUCCESS;
	}

	if (!strcmp(wdp->buffer->str[1],"*"))
	{
		WONKRNL_USER_CHANNEL *channel;
		LINKED_LIST_ITEM *item;

		for (item = listGetFirstItem(&dccExtension->user->channels);
				item && item->value;
				item = listGetNextItem(item))
		{
			channel = (WONKRNL_USER_CHANNEL *)item->value;

			if ((!dccCheckGlobalPermissions(dccExtension->user,WONKRNL_ACCESS_BANS)) 
				&& (!dccCheckChannelPermissions(dccExtension->user,channel->channel->channel,WONKRNL_ACCESS_BANS)))
			{
				wonDccSend(wdp->ctx,wdp->dcc,"No permissions for bans on \002%s\002.\n",channel->channel->channel);
				return WON_STATUS_SUCCESS;
			}

			if (wonkrnlBanDelete(channel->channel,wdp->buffer->str[2]) == WON_STATUS_SUCCESS)
			{
				wonIrcSend(wdp->ctx,dccExtension->irc,
							"MODE %s -b %s\n",
							channel->channel->channel,
							wdp->buffer->str[2]);
				wonDccSend(wdp->ctx,wdp->dcc,"Successfully deleted ban from channel \002%s\002.\n",channel->channel->channel);
			}
			else
				wonDccSend(wdp->ctx,wdp->dcc,"No such ban exists on \002%s\002.\n",channel->channel->channel);
		}
	}
	else
	{
		WONKRNL_USER_CHANNEL *channel = wonkrnlUserChannelHas(dccExtension->user,wdp->buffer->str[1]);

		if (!channel)
		{
			wonDccSend(wdp->ctx,wdp->dcc,"The channel \002%s\002 does not exist.\n",wdp->buffer->str[1]);
			return WON_STATUS_SUCCESS;
		}

		if ((!dccCheckGlobalPermissions(dccExtension->user,WONKRNL_ACCESS_BANS)) 
			&& (!dccCheckChannelPermissions(dccExtension->user,channel->channel->channel,WONKRNL_ACCESS_BANS)))
		{
			wonDccSend(wdp->ctx,wdp->dcc,"No.\n");
			return WON_STATUS_SUCCESS;
		}

		if (wonkrnlBanDelete(channel->channel,wdp->buffer->str[2]) == WON_STATUS_SUCCESS)
		{
			wonIrcSend(wdp->ctx,dccExtension->irc,
						"MODE %s -b %s\n",
						channel->channel->channel,
						wdp->buffer->str[2]);
			wonDccSend(wdp->ctx,wdp->dcc,"Successfully deleted ban from channel \002%s\002.\n",channel->channel->channel);
		}
		else
			wonDccSend(wdp->ctx,wdp->dcc,"No such ban exists on \002%s\002.\n",channel->channel->channel);
	}

	return WON_STATUS_SUCCESS;
}

uint32 dccBanList(WON_DCC_PACKET *wdp)
{
	WONKRNL_DCC_EXTENSION *dccExtension = (WONKRNL_DCC_EXTENSION *)wdp->dcc->dccExtension;
	WONKRNL_USER_CHANNEL *channel;
	LINKED_LIST_ITEM *item, *banItem;
	WONKRNL_BAN *ban;

	wonDccSend(wdp->ctx,wdp->dcc,"Listing ban list information...\n");

	if (wdp->buffer->words == 1)
	{
		for (item = listGetFirstItem(&dccExtension->user->channels);
				item && item->value;
				item = listGetNextItem(item))
		{
			channel = (WONKRNL_USER_CHANNEL *)item->value;

			if ((!dccCheckGlobalPermissions(dccExtension->user,WONKRNL_ACCESS_BANS)) 
				&& (!dccCheckChannelPermissions(dccExtension->user,channel->channel->channel,WONKRNL_ACCESS_BANS)))
				continue;

			wonDccSend(wdp->ctx,wdp->dcc,"\002%s\002 (%02i):\n",channel->channel->channel,channel->channel->bans.items);

			for (banItem = listGetFirstItem(&channel->channel->bans);
				banItem && banItem->value;
				banItem = listGetNextItem(banItem))
			{
				ban = (WONKRNL_BAN *)banItem->value;

				wonDccSend(wdp->ctx,wdp->dcc,"  \002%s\002  (%s)\n",ban->mask,ban->reason);
			}
		}
	}
	else
	{
		channel = wonkrnlUserChannelHas(dccExtension->user,wdp->buffer->str[1]);

		if (!channel)
		{
			wonDccSend(wdp->ctx,wdp->dcc,"The channel \002%s\002 does not exist or you do not have access.\n",wdp->buffer->str[1]);
			return WON_STATUS_SUCCESS;
		}

		if ((!dccCheckGlobalPermissions(dccExtension->user,WONKRNL_ACCESS_BANS)) 
			&& (!dccCheckChannelPermissions(dccExtension->user,channel->channel->channel,WONKRNL_ACCESS_BANS)))
			return WON_STATUS_FAILURE;

		wonDccSend(wdp->ctx,wdp->dcc,"\002%s\002 (%02i):\n",channel->channel->channel,channel->channel->bans.items);

		for (banItem = listGetFirstItem(&channel->channel->bans);
			banItem && banItem->value;
			banItem = listGetNextItem(banItem))
		{
			ban = (WONKRNL_BAN *)banItem->value;

			wonDccSend(wdp->ctx,wdp->dcc,"  \002%s\002  (%s)\n",ban->mask,ban->reason);
		}
	}

	return WON_STATUS_SUCCESS;
}

uint32 dccLoadModule(WON_DCC_PACKET *wdp)
{
	WONKRNL_DCC_EXTENSION *dccExtension = (WONKRNL_DCC_EXTENSION *)wdp->dcc->dccExtension;

	if (wdp->buffer->words < 3)
	{
		wonDccSend(wdp->ctx,wdp->dcc,"Usage: \037loadmodule\037 [\002modulepath\002] [\002modulename\002]\n");
		wonDccSend(wdp->ctx,wdp->dcc,"      Load a new module.\n");
		return WON_STATUS_SUCCESS;
	}

	if (!dccCheckGlobalPermissions(dccExtension->user,WONKRNL_ACCESS_MASTER))
		return WON_STATUS_SUCCESS;

	switch (wonModuleLoad(wdp->ctx,wdp->buffer->str[1],wdp->buffer->str[2]))
	{
		case WON_STATUS_EXISTS:
			wonDccSend(wdp->ctx,wdp->dcc,"The module \002%s\002 is already loaded.\n",wdp->buffer->str[2]);
			break;
		case WON_STATUS_FAILURE:
			wonDccSend(wdp->ctx,wdp->dcc,"The module \002%s\002 could NOT be loaded.\n",wdp->buffer->str[2]);
			break;
		default:
			wonDccSend(wdp->ctx,wdp->dcc,"Successfully loaded module \002%s\002.\n",wdp->buffer->str[2]);
			break;
	}

	return WON_STATUS_SUCCESS;
}

uint32 dccUnloadModule(WON_DCC_PACKET *wdp)
{
	WONKRNL_DCC_EXTENSION *dccExtension = (WONKRNL_DCC_EXTENSION *)wdp->dcc->dccExtension;

	if (wdp->buffer->words < 2)
	{
		wonDccSend(wdp->ctx,wdp->dcc,"Usage: \037unloadmodule\037 [\002modulename\002]\n");
		wonDccSend(wdp->ctx,wdp->dcc,"      Unload a module.\n");
		return WON_STATUS_SUCCESS;
	}

	if (!dccCheckGlobalPermissions(dccExtension->user,WONKRNL_ACCESS_MASTER))
		return WON_STATUS_SUCCESS;

	switch (wonModuleUnload(wdp->ctx,wdp->buffer->str[1]))
	{
		case WON_STATUS_NOEXISTS:
			wonDccSend(wdp->ctx,wdp->dcc,"The module \002%s\002 is not loaded.\n",wdp->buffer->str[1]);
			break;
		case WON_STATUS_FAILURE:
			wonDccSend(wdp->ctx,wdp->dcc,"The module \002%s\002 could NOT be unloaded.\n",wdp->buffer->str[1]);
			break;
		default:
			wonDccSend(wdp->ctx,wdp->dcc,"Successfully unloaded module \002%s\002.\n",wdp->buffer->str[1]);
			break;
	}


	return WON_STATUS_SUCCESS;
}

uint32 dccReloadModule(WON_DCC_PACKET *wdp)
{
	WONKRNL_DCC_EXTENSION *dccExtension = (WONKRNL_DCC_EXTENSION *)wdp->dcc->dccExtension;
	char modulePath[MAX_PATH];
	WON_MODULE_CTX *modctx;

	if (wdp->buffer->words < 2)
	{
		wonDccSend(wdp->ctx,wdp->dcc,"Usage: \037reloadmodule\037 [\002modulename\002]\n");
		wonDccSend(wdp->ctx,wdp->dcc,"      Reload a module.\n");
		return WON_STATUS_SUCCESS;
	}

	if (!dccCheckGlobalPermissions(dccExtension->user,WONKRNL_ACCESS_MASTER))
		return WON_STATUS_SUCCESS;

	if (!(modctx = (WON_MODULE_CTX *)hashStringKeyGetValue(&wdp->ctx->moduleHash,wdp->buffer->str[1])))
	{
		wonDccSend(wdp->ctx,wdp->dcc,"The module \002%s\002 is not loaded.\n",wdp->buffer->str[1]);
		return WON_STATUS_SUCCESS;
	}

	strncpy(modulePath,modctx->modulePath,MAX_PATH-1);

	switch (wonModuleUnload(wdp->ctx,wdp->buffer->str[1]))
	{
		case WON_STATUS_NOEXISTS:
			wonDccSend(wdp->ctx,wdp->dcc,"The module \002%s\002 is not loaded.\n",wdp->buffer->str[1]);
			break;
		case WON_STATUS_FAILURE:
			wonDccSend(wdp->ctx,wdp->dcc,"The module \002%s\002 could NOT be unloaded.\n",wdp->buffer->str[1]);
			break;
		default:
			wonDccSend(wdp->ctx,wdp->dcc,"Successfully unloaded module \002%s\002.\n",wdp->buffer->str[1]);
			break;
	}

	switch (wonModuleLoad(wdp->ctx,modulePath,wdp->buffer->str[1]))
	{
		case WON_STATUS_EXISTS:
			wonDccSend(wdp->ctx,wdp->dcc,"The module \002%s\002 is already loaded.\n",wdp->buffer->str[1]);
			break;
		case WON_STATUS_FAILURE:
			wonDccSend(wdp->ctx,wdp->dcc,"The module \002%s\002 could NOT be loaded.\n",wdp->buffer->str[1]);
			break;
		default:
			wonDccSend(wdp->ctx,wdp->dcc,"Successfully loaded module \002%s\002.\n",wdp->buffer->str[1]);
			break;
	}

	return WON_STATUS_SUCCESS;
}

uint32 dccListModules(WON_DCC_PACKET *wdp)
{
	LINKED_LIST_ITEM *item;
	WON_MODULE_CTX *modctx = NULL;

	wonDccSend(wdp->ctx,wdp->dcc,"Listing \037%u\037 modules...\n",wdp->ctx->moduleList.items);

	for (item = listGetFirstItem(&wdp->ctx->moduleList);
		item && item->value;
		item = listGetNextItem(item))
	{
		modctx = (WON_MODULE_CTX *)item->value;

		wonDccSend(wdp->ctx,wdp->dcc,"  \002%s\002 (\037%d KB\037)\n",modctx->moduleName,modctx->moduleSize);
	}

	return WON_STATUS_SUCCESS;
}

uint32 dccMsg(WON_DCC_PACKET *wdp)
{
	WONKRNL_DCC_EXTENSION *dccExtension = (WONKRNL_DCC_EXTENSION *)wdp->dcc->dccExtension;
	char str[512];

	if (wdp->buffer->words < 3)
	{
		wonDccSend(wdp->ctx,wdp->dcc,"Usage: \037msg\037 [\002nick\002] [\002text\002]\n");
		wonDccSend(wdp->ctx,wdp->dcc,"      Message a nickname specific text.\n");
		return WON_STATUS_SUCCESS;
	}

	if (!dccCheckGlobalPermissions(dccExtension->user,WONKRNL_ACCESS_MSGS))
	{
		wonDccSend(wdp->ctx,wdp->dcc,"No.\n");
		return WON_STATUS_SUCCESS;
	}

	if (!dccExtension->irc)
	{
		wonDccSend(wdp->ctx,wdp->dcc,"I'm not currently on IRC.\n");
		return WON_STATUS_SUCCESS;
	}

	str[511] = 0;
	cacheCompileWords(wdp->buffer,2,wdp->buffer->words," ",str,511);

	wonIrcSend(wdp->ctx,dccExtension->irc,
				"PRIVMSG %s :%s\n",
				wdp->buffer->str[1],
				str);

	wonDccSend(wdp->ctx,wdp->dcc,"(\037msg/%s\037) %s\n",wdp->buffer->str[1],str);

	return WON_STATUS_SUCCESS;
}

uint32 dccHelp(WON_DCC_PACKET *wdp)
{
	wonDccSend(wdp->ctx,wdp->dcc,"[wonkrnl level functions]\n");
	wonDccSend(wdp->ctx,wdp->dcc,"  \002User manipulation\002\n");
	wonDccSend(wdp->ctx,wdp->dcc,"   adduser deluser userinfo uattr cattr\n");
	wonDccSend(wdp->ctx,wdp->dcc,"   addhost delhost passwd boot\n");
	wonDccSend(wdp->ctx,wdp->dcc,"  \002Channel manipluation\002\n");
	wonDccSend(wdp->ctx,wdp->dcc,"   addchan delchan chaninfo active\n");
	wonDccSend(wdp->ctx,wdp->dcc,"   say op invite kick addban delban\n");
	wonDccSend(wdp->ctx,wdp->dcc,"   banlist\n");
	wonDccSend(wdp->ctx,wdp->dcc,"  \002Module manipulation\002\n");
	wonDccSend(wdp->ctx,wdp->dcc,"   loadmodule unloadmodule listmodules\n");
	wonDccSend(wdp->ctx,wdp->dcc,"   reloadmodule\n");
	wonDccSend(wdp->ctx,wdp->dcc,"  \002Misc functionality\002\n");
	wonDccSend(wdp->ctx,wdp->dcc,"   msg version help quit echo save who\n");

	return WON_STATUS_SUCCESS;
}

uint32 dccVersion(WON_DCC_PACKET *wdp)
{
	wonDccSend(wdp->ctx,wdp->dcc,"wonkrnl: %s\n",WONKRNL_VERSION);
	return WON_STATUS_SUCCESS;
}

uint32 dccQuit(WON_DCC_PACKET *wdp)
{
	wonDccSend(wdp->ctx,wdp->dcc,"keep your women close and your balls closer.\n");
	wonDccDeregisterConnection(wdp->ctx,wdp->dcc);

	return WON_STATUS_SUCCESS;
}

uint32 dccEcho(WON_DCC_PACKET *wdp)
{
	WONKRNL_DCC_EXTENSION *dccExtension = (WONKRNL_DCC_EXTENSION *)wdp->dcc->dccExtension;

	if (dccExtension->echo)
		wonDccSend(wdp->ctx,wdp->dcc,"echo is now off.\n");	
	else
		wonDccSend(wdp->ctx,wdp->dcc,"echo is now on.\n");

	dccExtension->echo = !dccExtension->echo;

	return WON_STATUS_SUCCESS;
}

uint32 dccSave(WON_DCC_PACKET *wdp)
{
	WONKRNL_DCC_EXTENSION *dccExtension = (WONKRNL_DCC_EXTENSION *)wdp->dcc->dccExtension;

	if (!dccCheckGlobalPermissions(dccExtension->user,WONKRNL_ACCESS_MASTER))
		return WON_STATUS_SUCCESS;

	switch (wonConfigurationSave(wdp->ctx))
	{
		case WON_STATUS_FAILURE:
			wonDccSend(wdp->ctx,wdp->dcc,"One or more modules failed to save configuration.\n");
			break;
		default:
			wonDccSend(wdp->ctx,wdp->dcc,"Successfully saved configuration.\n");
			break;
	}

	return WON_STATUS_SUCCESS;
}

uint32 dccWho(WON_DCC_PACKET *wdp)
{
	WONKRNL_DCC_EXTENSION *dccExtension;
	WON_DCC_CONNECTION *dcc;
	int x = 0;

	wonDccSend(wdp->ctx,wdp->dcc,"There are \037%lu\037 users connected.\n",wdp->ctx->dccConnections.items);

	while (wonDccEnum(wdp->ctx,x++,&dcc) != WON_STATUS_FAILURE)
	{
		dccExtension = (WONKRNL_DCC_EXTENSION *)dcc->dccExtension;

		if (!dccExtension || !dccExtension->user)
			continue;

		wonDccSend(wdp->ctx,wdp->dcc,
				"\002%10s\002 on network \037%s\037 (idle %lu seconds)\n",
				dccExtension->user->username,
				dccExtension->network->name,
				time(NULL) - dccExtension->idleTime);
	}

	return WON_STATUS_SUCCESS;
}

