#ifndef _WONKRNL_H
#define _WONKRNL_H

#define WONKRNL_VERSION		"built "  __TIME__ " " __DATE__ 
#define WONKRNL_DCC_STR		"[\037won\037]"

#define WONKRNL_EXT(ext) ((WONKRNL_MODULE_EXTENSION *)ext)

#define WONKRNL_STATE_CHANNEL_IN		(1 << 1)
#define WONKRNL_STATE_CHANNEL_OP		(1 << 2)
#define WONKRNL_STATE_CHANNEL_BANNED		(1 << 3)

#define WONKRNL_ACCESS_OPS			(1 << 1)
#define WONKRNL_ACCESS_AUTO_OPS			(1 << 2)
#define WONKRNL_ACCESS_MASTER			(1 << 3)
#define WONKRNL_ACCESS_INVITES			(1 << 4)
#define WONKRNL_ACCESS_KICKS			(1 << 5)
#define WONKRNL_ACCESS_BANS			(1 << 6)
#define WONKRNL_ACCESS_MSGS			(1 << 7)

typedef struct _wonkrnl_module_extension {
	void 		*currentNetwork;	/* Current Network during config 	*/
	
	LINKED_LIST	networks;
} WONKRNL_MODULE_EXTENSION;

typedef struct _wonknrl_network {
	char		*name;			/* Network name		*/
	char		*nickname;		/* Primary nickname	*/
	void		*currentChannel;	/* Current channel	*/
	void 		*currentUser;		/* Current user		*/

	unsigned long	vhost;			/* Vhost addr		*/

	char		*versionReply;		/* Version reply	*/
	char 		*dccUserPrompt;
	char		*dccPassPrompt;
	char		*motdFile;		
	
	LINKED_LIST 	channels;		
	LINKED_LIST	servers;		/* Servers		*/
	LINKED_LIST	users;			/* Users		*/
} WONKRNL_NETWORK;

typedef struct _wonkrnl_channel {
	WONKRNL_NETWORK	*network;

	unsigned int	state;

	char 		*channel;
	char		*key;

	LINKED_LIST	bans;			/* Bans			*/
} WONKRNL_CHANNEL;

typedef struct _wonkrnl_server {
	WONKRNL_NETWORK *network;

	unsigned long	addr;
	unsigned int	port;
} WONKRNL_SERVER;

typedef struct _wonkrnl_ban {
	WONKRNL_CHANNEL	*channel;

	char 		*mask;
	char		*reason;
} WONKRNL_BAN;

typedef struct _wonkrnl_user {
	WONKRNL_NETWORK *network;

	char		*username;		/* Username		*/
	unsigned char	password[16];		/* MD5 password digest	*/
	unsigned int	gaccess;		/* global access	*/

	LINKED_LIST	channels;		/* channel access list	*/
	LINKED_LIST	hosts;			/* hosts		*/
} WONKRNL_USER;

typedef struct _wonkrnl_user_channel {
	WONKRNL_USER	*user;
	WONKRNL_CHANNEL	*channel;

	unsigned int 	access;
} WONKRNL_USER_CHANNEL;

typedef struct _wonkrnl_dcc_extension {
	WONKRNL_NETWORK		*network;
	WONKRNL_USER		*user;
	WONKRNL_CHANNEL		*currentChannel;
	WON_IRC_CONNECTION	*irc;
	char			currentNickname[WON_MAX_NICKNAME_SIZE];
	uint32			loginTime;
	uint32			idleTime;

	char			echo;
} WONKRNL_DCC_EXTENSION;

/* Configuration handlers */

uint32 configStart(WON_CONFIG_PACKET *wcp);
uint32 configEnd(WON_CONFIG_PACKET *wcp);
uint32 configNickname(WON_CONFIG_PACKET *wcp);
uint32 configVhost(WON_CONFIG_PACKET *wcp);
uint32 configVersionReply(WON_CONFIG_PACKET *wcp);
uint32 configDccUserPrompt(WON_CONFIG_PACKET *wcp);
uint32 configDccPassPrompt(WON_CONFIG_PACKET *wcp);
uint32 configMotdFile(WON_CONFIG_PACKET *wcp);
uint32 configServer(WON_CONFIG_PACKET *wcp);
uint32 configBeginChannel(WON_CONFIG_PACKET *wcp);
uint32 configEndChannel(WON_CONFIG_PACKET *wcp);
uint32 configChannelKey(WON_CONFIG_PACKET *wcp);
uint32 configAddBan(WON_CONFIG_PACKET *wcp);

/* User config */

uint32 configBeginUser(WON_CONFIG_PACKET *wcp);
uint32 configEndUser(WON_CONFIG_PACKET *wcp);
uint32 configPassword(WON_CONFIG_PACKET *wcp);
uint32 configGlobalAccess(WON_CONFIG_PACKET *wcp);
uint32 configAllowHost(WON_CONFIG_PACKET *wcp);
uint32 configChannelAccess(WON_CONFIG_PACKET *wcp);

WONKRNL_CHANNEL *configLookupChannel(WON_IRC_CONNECTION *irc,const char *channel);
WONKRNL_USER	*configLookupUser(WON_IRC_CONNECTION *irc,const char *user);
WONKRNL_USER	*configLookupUserByMask(WON_IRC_CONNECTION *irc,const char *mask);

void configSave(WON_MODULE_CTX *modctx,FILE *fd);

/* Ioctl routines and defines */


/* Irc */

void ircConnectNetwork(WON_MODULE_CTX *modctx,WONKRNL_NETWORK *network);
void ircLinkUp(WON_MODULE_CTX *modctx,WON_IRC_CONNECTION *conn);
void ircLinkDown(WON_MODULE_CTX *modctx,WON_IRC_CONNECTION *conn);

/* Handlers */
uint32 ircConnected(WON_IRC_PACKET *wip);	/* 001		1 */
uint32 ircNamesList(WON_IRC_PACKET *wip);	/* 353		1 */
uint32 ircEndMotd(WON_IRC_PACKET *wip);		/* 376		1 */
uint32 ircNick(WON_IRC_PACKET *wip);		/* 433		1 */
uint32 ircInviteOnly(WON_IRC_PACKET *wip);	/* 473		1 */
uint32 ircRegister(WON_IRC_PACKET *wip);	/* 451 		1 */
uint32 ircPrivmsg(WON_IRC_PACKET *wip);		/* PRIVMSG	1 */
uint32 ircMode(WON_IRC_PACKET *wip);		/* MODE		1 */
uint32 ircJoin(WON_IRC_PACKET *wip);		/* JOIN		1 */
uint32 ircPart(WON_IRC_PACKET *wip);		/* PART		1 */
uint32 ircNickChange(WON_IRC_PACKET *wip);	/* NICK		1 */
uint32 ircKick(WON_IRC_PACKET *wip);		/* KICK		1 */
uint32 ircInvite(WON_IRC_PACKET *wip);		/* INVITE	1 */
uint32 ircPing(WON_IRC_PACKET *wip);		/* PING		0 */

/* DCC */

void dccLinkUp(WON_MODULE_CTX *modctx,WON_DCC_CONNECTION *conn);
void dccLinkDown(WON_MODULE_CTX *modctx,WON_DCC_CONNECTION *conn);

void dccEstablishConnection(WON_IRC_PACKET *wip);

/* Handlers */

uint32 dccLowlevel(WON_MODULE_CTX *modctx,WON_PACKET *packet);
uint32 dccCheckGlobalPermissions(WONKRNL_USER *user,uint32 level);
uint32 dccCheckChannelPermissions(WONKRNL_USER *user,const char *channel,uint32 level);
void dccSendMotd(WON_CTX *ctx,WON_DCC_CONNECTION *dcc, WONKRNL_DCC_EXTENSION *dccExtension);

uint32 dccAddUser(WON_DCC_PACKET *wdp);
uint32 dccDelUser(WON_DCC_PACKET *wdp);
uint32 dccUserInfo(WON_DCC_PACKET *wdp);
uint32 dccUattr(WON_DCC_PACKET *wdp);
uint32 dccCattr(WON_DCC_PACKET *wdp);
uint32 dccAddHost(WON_DCC_PACKET *wdp);
uint32 dccDelHost(WON_DCC_PACKET *wdp);
uint32 dccPasswd(WON_DCC_PACKET *wdp);
uint32 dccBoot(WON_DCC_PACKET *wdp);

uint32 dccAddChan(WON_DCC_PACKET *wdp);
uint32 dccDelChan(WON_DCC_PACKET *wdp);
uint32 dccChanInfo(WON_DCC_PACKET *wdp);
uint32 dccConsole(WON_DCC_PACKET *wdp);
uint32 dccSay(WON_DCC_PACKET *wdp);
uint32 dccOp(WON_DCC_PACKET *wdp);
uint32 dccInvite(WON_DCC_PACKET *wdp);
uint32 dccKick(WON_DCC_PACKET *wdp);
uint32 dccAddBan(WON_DCC_PACKET *wdp);
uint32 dccDelBan(WON_DCC_PACKET *wdp);
uint32 dccBanList(WON_DCC_PACKET *wdp);

uint32 dccLoadModule(WON_DCC_PACKET *wdp);
uint32 dccUnloadModule(WON_DCC_PACKET *wdp);
uint32 dccReloadModule(WON_DCC_PACKET *wdp);
uint32 dccListModules(WON_DCC_PACKET *wdp);

uint32 dccMsg(WON_DCC_PACKET *wdp);
uint32 dccHelp(WON_DCC_PACKET *wdp);
uint32 dccVersion(WON_DCC_PACKET *wdp);
uint32 dccQuit(WON_DCC_PACKET *wdp);
uint32 dccEcho(WON_DCC_PACKET *wdp);
uint32 dccSave(WON_DCC_PACKET *wdp);
uint32 dccWho(WON_DCC_PACKET *wdp);

/* Helper functions */

uint32 wonkrnlResolveAddress(const char *address);

WONKRNL_NETWORK *wonkrnlNetworkAdd(WONKRNL_MODULE_EXTENSION *krnl,const char *network);
WONKRNL_NETWORK *wonkrnlNetworkHas(WONKRNL_MODULE_EXTENSION *krnl,const char *network);
uint32 wonkrnlNetworkDelete(WONKRNL_MODULE_EXTENSION *krnl,const char *network);

WONKRNL_SERVER *wonkrnlServerAdd(WONKRNL_NETWORK *network,const char *server,unsigned int port);
WONKRNL_SERVER *wonkrnlServerHas(WONKRNL_NETWORK *network,const char *server,unsigned int port);
uint32 wonkrnlServerDelete(WONKRNL_NETWORK *network,const char *server,unsigned int port);

WONKRNL_USER *wonkrnlUserAdd(WONKRNL_NETWORK *network,const char *username);
WONKRNL_USER *wonkrnlUserHas(WONKRNL_NETWORK *network,const char *username);
uint32 wonkrnlUserDelete(WONKRNL_NETWORK *network,const char *username);

WONKRNL_CHANNEL *wonkrnlChannelAdd(WONKRNL_NETWORK *network,const char *channel);
WONKRNL_CHANNEL *wonkrnlChannelHas(WONKRNL_NETWORK *network,const char *channel);
uint32 wonkrnlChannelDelete(WONKRNL_NETWORK *network,const char *channel);

WONKRNL_BAN *wonkrnlBanAdd(WONKRNL_CHANNEL *channel,const char *mask,const char *reason);
WONKRNL_BAN *wonkrnlBanHas(WONKRNL_CHANNEL *channel,const char *mask);
uint32 wonkrnlBanDelete(WONKRNL_CHANNEL *channel,const char *mask);

WONKRNL_USER_CHANNEL *wonkrnlUserChannelAdd(WONKRNL_USER *user,const char *channel);
WONKRNL_USER_CHANNEL *wonkrnlUserChannelHas(WONKRNL_USER *user,const char *channel);
uint32 wonkrnlUserChannelDelete(WONKRNL_USER *user,const char *channel);

const char *wonkrnlUserHostAdd(WONKRNL_USER *user,const char *host);
const char *wonkrnlUserHostHas(WONKRNL_USER *user,const char *host);
uint32 wonkrnlUserHostDelete(WONKRNL_USER *user,const char *host);

#endif
