/* 
 * Wonton
 *
 * Afro Productions - England Blue Division (c)
 * 
 * Authors:  Matt Miller   (mmiller@hick.org)
 *           Brett Spedale (soap@technique.org)
 *
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>

/*
   * These defines affect the wild carding for the function that covers 
   * matching later in the file. If you wish to change the characters
   * for it to consider matches of a certain sort, then here they are.
*/

#define WILDCARD_OFF '\\'
#define WILDCARD_ALL '*'
#define WILDCARD_NUMBER '%'
#define WILDCARD_CHAR '?'
#define WILDCARD_SPACE '~'

#define MATCH (iMatch+iSaved+iSofar)

void fnCutItUp(char *buf, char *string, char *begin, char *end);
void fnSearchAndReplace(char *buf, char *string, char search, char replace);
void fnDivideString (char *first, char *rest, char divider);
int fnReadFile (char *filename, void *tocall);
int fnWildCardMatch (register unsigned char *pFistString, register unsigned char *pSecondString);

/*
   * FUNCTION
   * void fnCutItUp(char *buf, char *string, char *begin, char *end)
   *
   * This function puts every character of string in buf, starting at
   * begin and ending at end.
   *
   * Arguments:
   *	buf: the buffer for the result to be stored in.
   *	string: the string to copy.
   *	begin: the string to begin at.
   *	end: the string to end at.
   *
   *	NOTE: This function strickly copies it what you tell it to.
   *	For example, if the string is "blah1 blah2 blah3", and you
   *	tell it to copy from blah1 to blah3, it will include the
   *	trailing and ending whitespaces on blah2, thus " blah 2".
   *	Also, don't be afraid to use this function when you arn't
   *	sure that the string doesn't end in the ending. If it doesn't,
   *	the whole string after begin will simply be copied.
   *
   * Returns: NULL
   *
*/

void fnCutItUp(char *buf, char *string, char *begin, char *end)
{
   char *result;
   char *ending;
   char *buffer;

   int i, buflen;

   result = (char *) malloc(sizeof(buf)+1);

   buffer = strstr(string, begin);

   for (i=0;i!=strlen(begin);i++)
   {
      buffer++;
   }

   i = 0;

   ending = strstr(buffer, end);
   buflen = ending-buffer;

   if (0 > buflen)
   {
      sprintf(result, "%s", buffer);
   }
   else
   {
      while (i < buflen)
      {
	 result[i] = *buffer;
	 buffer++;
	 i++;
      }

      result[i] = '\0';
   }
   sprintf(buf, "%s", result);
   free(result);
}

/*
   * FUNCTION
   * void fnSearchAndReplace(char *buf, char *string, char search, char replace)
   *
   * Search for a character in string and replace with a character, and store
   * the result in the given buffer. I have found this to be a very useful
   * function.
   *
   * Arguments:
   *	buf: the buffer to store the result in.
   *	string: the string to search.
   *	search: the character to search for.
   *	replace: the character to replace search with.
   *
   * Returns: NULL
   *
*/

void fnSearchAndReplace(char *buf, char *string, char search, char replace)
{
   char *ptr;
   char *result;
   int i;

   result = (char *) malloc(strlen(string)+1);

/* fill the string with NULLs.. this isn't really needed */

   for (i=0;i!=(strlen(string)+1);i++)
   {
      result[i] = '\0';
   }

   i = 0;

   ptr = string;

   while (*ptr)
   {
      if (*ptr == search)
      {
	 result[i] = replace;
      }
      else
      {
	 result[i] = (char) *ptr;
      }

      ptr++;
      i++;
   }

   sprintf(buf, "%s", result);
   free(result);
}

/*
   * FUNCTION
   * void fnDivideString(char *first, char *rest, char divider);
   *
   * Split the first word off of a string and put it in first. Useful for
   * manipulating umasks, regular whitespaced text, ect.
   *
   * Arguments:
   *      first: a pointer to where the resulting string should be stored.
   *      rest: the "rest" of the string. hence the name =)
   *      divider: the dividing character.. such as a space, period, ect.
   *
   * Returns: NULL
*/

void fnDivideString (char *first, char *rest, char divider)
{
   char *ptr;

   ptr = strchr (rest, divider);

   if (ptr == NULL)
   {
      if ((first != rest) && (first != NULL))
	 first[0] = 0;

      return;
   }

   *ptr = 0;

   if (first != NULL)
      strcpy (first, rest);
   if (first != rest)
      strcpy (rest, ptr + 1);
}

/*
   * FUNCTION
   * int fnReadFile(char *filename, void *tocall)
   *
   * Used to read a file and pass the contents to a function. I can't
   * think of a practical use for it but their has got to be one =)
   *
   * Arguments:
   *      filename: the filename to read
   *      tocall: the function to call
   *
   *      NOTE: the tocall should just be something like "fnDisplayStuff".
   *      It is intalized as a real function inside the fnReadFile function.
   *
   * Returns: 1 on sucess, 0 on a failure.
*/

int fnReadFile (char *filename, void *tocall)
{
   char buf[256];
   void (*function) (char *);
   FILE *fp;

   function = tocall;

   if ((fp = fopen (filename, "r")) == NULL)
   {
      exit (-1);
      return 0;
   }
   while (fgets (buf, 256, fp) != NULL)
   {
      function (buf);
   }

   fclose (fp);

   return 1;
}

/*
   * FUNCTION
   * int fnWildCardMatch (register unsigned char *pFirstString,
   *	 register unsigned char *pSecondString)
   *
   * Used to see if one string matches another. Wildcards are excepted.
   * 
   * Arguments:
   *	pFirstString: the first string
   *	pSecondString: the second string
   *
   * Returns: 1 if strings do match, 0 if they do not.
*/


int wonWildcardMatch(unsigned char *pFistString,
   			unsigned char *pSecondString)
{
   unsigned char *ma = pFistString;
   unsigned char *upLsm = NULL;
   unsigned char *upLsn = NULL;
   unsigned char *upLpm = NULL;
   unsigned char *upLpn = NULL;

   int 		iMatch = 1;
   int 		iSaved = 0;
   int 		iSpace = 0;

   register unsigned int iSofar = 0;

   if ((pFistString == 0) || (pSecondString == 0) || (!*pSecondString))
   {
      return 0;
   }

   while (*pSecondString)
   {

      if (*pFistString == WILDCARD_SPACE)
      {
	 iSpace = 0;

	 do
	 {
	    pFistString++;
	    iSpace++;
	 }

	 while ((*pFistString == WILDCARD_SPACE) || (*pFistString == ' '));
	 iSofar += iSpace;

	 while (*pSecondString == ' ')
	 {
	    pSecondString++;
	    iSpace--;
	 }

	 if (iSpace <= 0)
	    continue;
      }


      else
      {
	 switch (*pFistString)
	 {
	    case 0:
	       do
		  pFistString--;
	       while ((pFistString > ma) && (*pFistString == '?'));
	       if ((pFistString > ma) ? ((*pFistString == '*') &&
		  (pFistString[-1] != WILDCARD_OFF)) : (*pFistString == '*'))
	       {
		  return MATCH;
	       }

	       break;
	    case WILDCARD_NUMBER:
	       while (*(++pFistString) == WILDCARD_NUMBER);	/* Zap redundant %s      */
	       if (*pFistString != WILDCARD_ALL)
	       {
		  if (*pSecondString != ' ')
		  {
		     upLpm = pFistString;
		     upLpn = pSecondString;
		     iSaved += iSofar;

		     iSofar = 0;
		  }
		  continue;
	       }

	    case WILDCARD_ALL:
	       do
		  pFistString++;
	       while ((*pFistString == WILDCARD_ALL) || (*pFistString == WILDCARD_NUMBER));
	       upLsm = pFistString;
	       upLsn = pSecondString;
	       upLpm = 0;
	       iMatch += (iSaved + iSofar);	/* Save tally count      */
	       iSaved = iSofar = 0;
	       continue;

	    case WILDCARD_CHAR:
	       pFistString++;
	       pSecondString++;
	       continue;
	    case WILDCARD_OFF:
	       pFistString++;
	 }

	 if (tolower (*pFistString) == tolower (*pSecondString))
	 {
	    pFistString++;
	    pSecondString++;
	    iSofar++;
	    continue;
	 }
      }
      if (upLpm)
      {
	 pSecondString = ++upLpn;
	 pFistString = upLpm;
	 iSofar = 0;
	 if ((*pSecondString | 32) == 32)
	    upLpm = 0;
	 continue;
      }
      if (upLsm)
      {
	 pSecondString = ++upLsn;
	 pFistString = upLsm;

	 iSaved = iSofar = 0;
	 continue;
      }

      return 0;
   }

   while ((*pFistString == WILDCARD_ALL) || (*pFistString == WILDCARD_NUMBER))
      pFistString++;
   return (*pFistString) ? 0 : MATCH;
}
