#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <unistd.h>
#include <signal.h>

#include "won.h"
#include "wonloader.h"

void wonPrintVersion();
void wonInitialize(WON_CTX *ctx);
void wonDeinitialize(WON_CTX *ctx);

int main(int argc, char **argv)
{
	WON_CTX ctx;
	int c, noFork = 0;

	srand(time(NULL));

	wonInitialize(&ctx);

	while ((c = getopt(argc,argv,"c:vn")) != EOF)
	{
		switch (c)
		{
			case 'c':
				ctx.configurationFile = optarg;
				break;
			case 'v':
				wonPrintVersion();
				return 0;
			case 'n':
				noFork = 1;
				break;
		}
	}

	wonLoaderConfigurationParse(&ctx);

	if (!noFork)
	{
		ctx.pid = fork();

		if (ctx.pid == -1)
			fprintf(stdout,"wonloader: WARNING: fork failed.\n");
		else if (ctx.pid != 0)
		{
			wonLog(&ctx,NULL,"forked into PID %lu.",ctx.pid);
			fprintf(stdout,"wonloader: successfully forked.  pid %lu.\n",ctx.pid);
			exit(1);
		}
	}

	wonLoaderExecute(&ctx);

	wonDeinitialize(&ctx);

	return 1;
}

void wonPrintVersion()
{
	fprintf(stdout,"wonton: loader version 0.00 build 0\n");
	return;
}

void wonInitialize(WON_CTX *ctx)
{
	memset(ctx,0,sizeof(WON_CTX));

	signal(SIGPIPE,SIG_IGN);

	ctx->startTime		= time(NULL);
	ctx->configurationFile	= DEFAULT_CONFIGURATION_FILE;

	listInitialize(&ctx->moduleList,LIST_FLAG_FREEVALUE|LIST_FLAG_APPENDEND);
	listInitialize(&ctx->ircConnections,LIST_FLAG_FREEVALUE);
	listInitialize(&ctx->dccConnections,LIST_FLAG_FREEVALUE);
	listInitialize(&ctx->schedulerList,LIST_FLAG_FREEVALUE);
	listInitialize(&ctx->lowlevelRecvHandlers,0);
	listInitialize(&ctx->lowlevelSendHandlers,0);

	hashInitialize(&ctx->moduleHash,HASH_FLAG_ISDYNAMIC , 2048);	/* list frees value */
	hashInitialize(&ctx->ircHandlers,HASH_FLAG_ISDYNAMIC | HASH_FLAG_FREEVALUE, 2048);
	hashInitialize(&ctx->dccHandlers,HASH_FLAG_ISDYNAMIC | HASH_FLAG_FREEVALUE , 2048);
	hashInitialize(&ctx->configHandlers,HASH_FLAG_ISDYNAMIC | HASH_FLAG_FREEVALUE, 2048);

	ctx->initialized 	= 1;

	wonConfigRegisterLayeredHandler(ctx,
					NULL,
					WON_LAYER_FIRST,
					"load_module",
					wonLoaderConfigurationLoadModule);

	wonConfigRegisterLayeredHandler(ctx,
					NULL,
					WON_LAYER_FIRST,
					"log_file",
					wonLoaderConfigurationLogFile);
}

void wonDeinitialize(WON_CTX *ctx)
{
	if (!ctx->initialized)
		return;

	listDeinitialize(&ctx->moduleList);
	listDeinitialize(&ctx->ircConnections);
	listDeinitialize(&ctx->dccConnections);
	listDeinitialize(&ctx->schedulerList);
	listDeinitialize(&ctx->lowlevelRecvHandlers);
	listDeinitialize(&ctx->lowlevelSendHandlers);

	hashDeinitialize(&ctx->moduleHash);
	hashDeinitialize(&ctx->ircHandlers);
	hashDeinitialize(&ctx->dccHandlers);
	hashDeinitialize(&ctx->configHandlers);

	wonConfigDeregisterLayeredHandler(ctx,
					 NULL,
					 "load_module",
					 wonLoaderConfigurationLoadModule);
	
	wonConfigDeregisterLayeredHandler(ctx,
					 NULL,
					 "log_file",
					 wonLoaderConfigurationLogFile);
}

